import functools
import logging

import pytest

from leapp.libraries.actor import setuptargetrepos_repomap
from leapp.libraries.actor.setuptargetrepos_repomap import get_default_repository_channels, RepoMapDataHandler
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import PESIDRepositoryEntry, RepoMapEntry, RepositoriesMapping


def make_pesid_repo(pesid, major_version, repoid, arch='x86_64', repo_type='rpm', channel='ga', rhui=''):
    """
    PESIDRepositoryEntry factory function allowing shorter data description in tests by providing default values.
    """
    return PESIDRepositoryEntry(
        pesid=pesid,
        major_version=major_version,
        repoid=repoid,
        arch=arch,
        repo_type=repo_type,
        channel=channel,
        rhui=rhui
    )


@pytest.fixture
def repomap_data_for_pesid_repo_retrieval():
    # NOTE: do not change order or repositories; particular unit-tests tests
    # use specific repositories for the comparison, expecting them on the right
    # position in the list
    repomap_data = RepositoriesMapping(
        mapping=[
            RepoMapEntry(source='pesid1', target=['pesid3', 'pesid2']),
            RepoMapEntry(source='pesid4', target=['pesid4']),
        ],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid'),
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-eus', channel='eus'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid'),
            make_pesid_repo('pesid3', '8', 'pesid3-repoid'),
            make_pesid_repo('pesid3', '8', 'pesid3-repoid-eus', channel='eus'),
            make_pesid_repo('pesid3', '8', 'pesid3-repoid-aws', rhui='aws'),
            make_pesid_repo('pesid4', '7', 'pesid4-repoid1'),
            make_pesid_repo('pesid4', '8', 'pesid4-repoid2'),
        ]
    )
    return repomap_data


def test_get_pesid_repo_entry(monkeypatch, repomap_data_for_pesid_repo_retrieval):
    """
    Test for the RepoMapDataHandler.get_pesid_repo_entry method.

    Verifies that the method correctly retrieves PESIDRepositoryEntry that are matching the OS major version
    and repoid.
    """
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    repositories = repomap_data_for_pesid_repo_retrieval.repositories
    handler = RepoMapDataHandler(repomap_data_for_pesid_repo_retrieval)

    fail_description = (
        'get_pesid_repo_entry method failed to find correct pesid repository that matches given parameters.')
    for exp_repo in repositories:
        result_repo = handler.get_pesid_repo_entry(exp_repo.repoid, exp_repo.major_version)
        assert result_repo == exp_repo, fail_description

    fail_description = (
        'get_pesid_repo_entry method found a pesid repository, but no repository should match given parameters.')
    assert handler.get_pesid_repo_entry('pesid1-repoid', '6') is None, fail_description
    assert handler.get_pesid_repo_entry('pesid1-repoid', '8') is None, fail_description
    assert handler.get_pesid_repo_entry('pesid1-repoid', '9') is None, fail_description
    assert handler.get_pesid_repo_entry('nonexisting-repo', '7') is None, fail_description


def test_get_target_pesids(monkeypatch, repomap_data_for_pesid_repo_retrieval):
    """
    Test for the RepoMapDataHandler.get_target_pesids method.

    Verifies that the method correctly tells what target pesids is the given source pesid mapped to.
    """
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    handler = RepoMapDataHandler(repomap_data_for_pesid_repo_retrieval)

    expected_target_pesids = ['pesid2', 'pesid3']
    actual_target_pesids = handler.get_target_pesids('pesid1')

    fail_description = (
        'The get_target_pesids method did not correctly identify what is the given source pesid mapped to.')
    assert expected_target_pesids == actual_target_pesids, fail_description

    fail_description = (
        'The get_target_pesids method found target pesids even if the source repository is not mapped.')
    assert [] == handler.get_target_pesids('pesid2'), fail_description
    assert [] == handler.get_target_pesids('pesid_no_mapping'), fail_description


def test_get_pesid_repos(monkeypatch, repomap_data_for_pesid_repo_retrieval):
    """
    Test for the RepoMapDataHandler.get_pesid_repos method.

    Verifies that the method is able to collect all PESIDRepositoryEntry present in the repomap data that
    match the given OS major version and the given pesid.
    """
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    handler = RepoMapDataHandler(repomap_data_for_pesid_repo_retrieval)
    repositories = repomap_data_for_pesid_repo_retrieval.repositories

    actual_pesid_repos = handler.get_pesid_repos('pesid3', '8')
    expected_pesid_repos = [repositories[3], repositories[4], repositories[5]]
    fail_description = 'The get_pesid_repos failed to find pesid repos matching the given criteria.'
    assert len(expected_pesid_repos) == len(actual_pesid_repos), fail_description
    for actual_pesid_repo in actual_pesid_repos:
        assert actual_pesid_repo in expected_pesid_repos, fail_description

    actual_pesid_repos = handler.get_pesid_repos('pesid1', '7')
    expected_pesid_repos = [repositories[0], repositories[1]]
    assert len(expected_pesid_repos) == len(actual_pesid_repos), fail_description
    for actual_pesid_repo in actual_pesid_repos:
        assert actual_pesid_repo in expected_pesid_repos, fail_description

    fail_description = (
        'The get_pesid_repos found some pesid repositories matching criteria, but there are no such repositories.')
    assert [] == handler.get_pesid_repos('pesid3', '7'), fail_description
    assert [] == handler.get_pesid_repos('pesid1', '8'), fail_description
    assert [] == handler.get_pesid_repos('nonexisting_pesid', '7'), fail_description


def test_get_source_pesid_repos(monkeypatch, repomap_data_for_pesid_repo_retrieval):
    """
    Test for the RepoMapDataHandler.get_source_pesid_repos method.

    Verifies that the method is able to collect all PESIDRepositoryEntry that match the given PES ID and
    have the major version same as the source system.
    """
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    handler = RepoMapDataHandler(repomap_data_for_pesid_repo_retrieval)
    repositories = repomap_data_for_pesid_repo_retrieval.repositories

    fail_description = (
        'The get_source_pesid_repos method failed to retrieve all pesid repos that match given pesid '
        'and have the major version same as the source system.')
    expected_pesid_repos = [repositories[0], repositories[1]]
    actual_pesid_repos = handler.get_source_pesid_repos('pesid1')
    assert len(expected_pesid_repos) == len(actual_pesid_repos), fail_description
    for actual_pesid_repo in actual_pesid_repos:
        assert actual_pesid_repo in expected_pesid_repos, fail_description

    fail_description = (
        'The get_source_pesid_repos method does not take into account the source system version correctly.'
    )
    monkeypatch.setattr(setuptargetrepos_repomap, 'get_source_major_version', lambda: '10')

    # Repeat the same test as above to make sure it respects the source OS major version
    assert [] == handler.get_source_pesid_repos('pesid1'), fail_description

    assert [] == handler.get_source_pesid_repos('pesid2'), fail_description
    assert [] == handler.get_source_pesid_repos('nonexisting_pesid'), fail_description


def test_get_target_pesid_repos(monkeypatch, repomap_data_for_pesid_repo_retrieval):
    """
    Test for the RepoMapDataHandler.get_target_pesid_repos method.

    Verifies that the method is able to collect all PESIDRepositoryEntry that match the given PES ID and
    have the major version same as the source system.
    """
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    handler = RepoMapDataHandler(repomap_data_for_pesid_repo_retrieval)
    repositories = repomap_data_for_pesid_repo_retrieval.repositories

    fail_description = (
        'The get_target_pesid_repos method failed to retrieve all pesid repos that match given pesid '
        'and have the major version same as the target system.')
    expected_pesid_repos = [repositories[3], repositories[4], repositories[5]]
    actual_pesid_repos = handler.get_target_pesid_repos('pesid3')
    assert len(expected_pesid_repos) == len(actual_pesid_repos), fail_description
    for actual_pesid_repo in actual_pesid_repos:
        assert actual_pesid_repo in expected_pesid_repos, fail_description

    fail_description = (
        'The get_target_pesid_repos method doesn\'t take into account the target system version correctly.'
    )
    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='9.4', dst_ver='10.0'))

    # Repeat the same test as above to make sure it respects the target OS major version
    assert [] == handler.get_target_pesid_repos('pesid3'), fail_description

    assert [] == handler.get_target_pesid_repos('pesid1'), fail_description
    assert [] == handler.get_target_pesid_repos('nonexisting_pesid'), fail_description


@pytest.fixture
def mapping_data_for_find_repository_equiv():
    repositories = [
        make_pesid_repo('pesid1', '7', 'pesid1-repoid'),
        make_pesid_repo('pesid1', '7', 'pesid1-repoid', channel='e4s', rhui='aws'),
        make_pesid_repo('pesid2', '8', 'pesid2-repoid1'),
        make_pesid_repo('pesid2', '8', 'pesid2-repoid2-s390x', arch='s390x'),
        # This repository is a better candidate than the full match equivalent, but _find_repository_target_equivalent
        # should not take into account channel priorities
        make_pesid_repo('pesid2', '8', 'pesid2-repoid2', arch='x86_64', channel='eus'),
        make_pesid_repo('pesid2', '8', 'pesid2-repoid3-srpm', repo_type='srpm'),
        make_pesid_repo('pesid2', '8', 'pesid2-repoid4.1', rhui='aws'),
        make_pesid_repo('pesid2', '8', 'pesid2-repoid4.2', channel='eus', rhui='aws'),
        make_pesid_repo('pesid3', '8', 'pesid3-repoid')
    ]
    return RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2', 'pesid3'])],
        repositories=repositories
    )


def test_find_repository_target_equivalent_fullmatch(monkeypatch, mapping_data_for_find_repository_equiv):
    """
    Test for the RepoMapDataHandler._find_repository_target_equivalent method.

    Verifies that the method can find the target equivalent for a repository that matches the source
    pesid repo parameters exactly when such repository is available in the repository mapping data.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    handler = RepoMapDataHandler(mapping_data_for_find_repository_equiv)

    expected_target_equivalent = mapping_data_for_find_repository_equiv.repositories[2]
    actual_target_equivalent = handler._find_repository_target_equivalent(
        mapping_data_for_find_repository_equiv.repositories[0], 'pesid2')

    fail_description = (
        'The _find_repository_target_equivalent failed to find equivalent that exactly matched the source pesid repo '
        'when there is such equivalent available in the repository mapping data.')
    assert expected_target_equivalent == actual_target_equivalent, fail_description


def test_find_repository_target_equivalent_fallback_to_default(monkeypatch,
                                                               mapping_data_for_find_repository_equiv):
    """
    Test for the RepoMapDataHandler._find_repository_target_equivalent method.

    Verifies that the method will find a target equivalent with matching some of the fallback
    channels if a target equivalent that matches the source pesid repository completely is not
    available in the repository mapping data.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    handler = RepoMapDataHandler(mapping_data_for_find_repository_equiv)
    repositories = mapping_data_for_find_repository_equiv.repositories

    fail_description = (
        'The _find_repository_target_equivalent failed to find repository with some of the fallback channels.')
    expected_target_equivalent = repositories[6]
    actual_target_equivalent = handler._find_repository_target_equivalent(repositories[1], 'pesid2')
    assert expected_target_equivalent == actual_target_equivalent, fail_description

    handler.set_default_channels(['eus', 'ga'])

    expected_target_equivalent = repositories[7]
    actual_target_equivalent = handler._find_repository_target_equivalent(repositories[1], 'pesid2')
    assert expected_target_equivalent == actual_target_equivalent, fail_description


def test_find_repository_target_equivalent_no_target_equivalent(monkeypatch,
                                                                mapping_data_for_find_repository_equiv):
    """
    Test for the RepoMapDataHandler._find_repository_target_equivalent method.

    Verifies that the  does not crash when there is no target repository that is equivalent to the
    source repository.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='s390x', src_ver='7.9', dst_ver='8.4'))
    handler = RepoMapDataHandler(mapping_data_for_find_repository_equiv)

    fail_description = 'The _find_repository_target_equivalent found target equivalent when there are none.'

    repository_with_no_equivalent = make_pesid_repo('pesid1', '7', 'pesid1-some-repoid', arch='s390x', rhui='aws')
    target_equivalent = handler._find_repository_target_equivalent(repository_with_no_equivalent, 'pesid2')
    assert target_equivalent is None, fail_description


def test_get_mapped_target_pesid_repos(monkeypatch, mapping_data_for_find_repository_equiv):
    """
    Test for the RepoMapDataHandler.get_mapped_target_pesid_repos method.

    Verifies that the method correctly builds a map mapping the target pesid to the best candidate
    pesid repos.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    repositories = mapping_data_for_find_repository_equiv.repositories

    handler = RepoMapDataHandler(mapping_data_for_find_repository_equiv)

    # Both pesid2 and pesid3 have an equivalent for provided source pesid repository
    fail_description = (
        'The get_mapped_target_pesid_repos failed to build a map mapping the target pesid '
        'to the best pesid repository candidate.')
    target_pesid_repos_map = handler.get_mapped_target_pesid_repos(repositories[0])
    expected_pesid_to_best_candidate_map = {'pesid2': repositories[2], 'pesid3': repositories[8]}
    assert target_pesid_repos_map == expected_pesid_to_best_candidate_map, fail_description

    # The pesid3 does not have an equivalent for provided source pesid repository (due to not having any rhui repos)
    target_pesid_repos_map = handler.get_mapped_target_pesid_repos(repositories[1])
    expected_pesid_to_best_candidate_map = {'pesid2': repositories[6], 'pesid3': None}
    assert target_pesid_repos_map == expected_pesid_to_best_candidate_map, fail_description


def test_get_mapped_target_repoids(monkeypatch, mapping_data_for_find_repository_equiv):
    """
    Test for the RepoMapDataHandler.get_mapped_target_repoids method.

    Verifies that the method returns a correct list of repoids that should be present on the target system.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    handler = RepoMapDataHandler(mapping_data_for_find_repository_equiv)
    repositories = mapping_data_for_find_repository_equiv.repositories

    fail_description = (
        'The get_mapped_target_repoids method failed to return the correct list of repoids ',
        'to be enabled on the target system.')
    # Both pesid2 and pesid3 have an equivalent for provided source pesid repository
    actual_target_repoids = handler.get_mapped_target_repoids(repositories[0])
    expected_target_repoids = {repositories[2].repoid, repositories[8].repoid}
    assert len(actual_target_repoids) == len(expected_target_repoids), fail_description
    assert set(actual_target_repoids) == expected_target_repoids, fail_description

    # The pesid3 does not have an equivalent for provided source pesid
    # repository -> only one repository should be in the produced list
    actual_target_repoids = handler.get_mapped_target_repoids(repositories[1])
    assert len(actual_target_repoids) == 1
    assert 'pesid2-repoid4.1' in actual_target_repoids, fail_description


def test_get_expected_target_repoids_simple(monkeypatch):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the method is able to produce a correct map that maps target pesid to the best
    candidate pesid repository when there is only one repoid enabled and the corresponding source
    pesid repository has exact match target equivalent.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid')
        ]
    )
    fail_description = 'Failed to get_expected_target_repoids with only one repository enabled on the source system.'
    handler = RepoMapDataHandler(repositories_mapping)
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid'])

    assert {'pesid2': repositories_mapping.repositories[1]} == target_repoids, fail_description


def test_get_expected_target_repoids_best_candidate_produced(monkeypatch):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the method is able to produce a correct map that maps target pesid to the best
    candidate pesid repository when there are two repositories with different priority channels
    belonging to the same pesid family enabled on the source system and both have target
    equivalents.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-eus', channel='eus'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-eus', channel='eus'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-e4s', channel='e4s'),
        ]
    )

    handler = RepoMapDataHandler(repositories_mapping)
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-eus'])

    fail_description = (
        'The get_expected_target_repoids failed to map target pesid to a pesid repository'
        'with the highest priority channel.'
    )
    assert {'pesid2': repositories_mapping.repositories[3]} == target_repoids, fail_description


def test_get_expected_target_repoids_fallback(monkeypatch):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the RepoMapDataHandler.get_expected_target_repoids method is able to produce a correct
    map that maps target pesid to the best candidate pesid repository when there is a repository
    on the source system that does not have exact match equivalent and some other with a fallback channel
    must be found.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-eus', channel='eus'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga'),
            make_pesid_repo('pesid1', '8', 'pesid2-repoid-e4s', channel='e4s'),
        ]
    )

    fail_description = (
        'The get_expected_target_repoids failed to find repository with a failback channel '
        'since there were no exact target equivalents.')

    handler = RepoMapDataHandler(repositories_mapping)
    handler.set_default_channels(['ga'])
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-eus'])

    assert {'pesid2': repositories_mapping.repositories[1]} == target_repoids, fail_description


def test_get_expected_target_pesid_repos_multiple_repositories(monkeypatch):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the RepoMapDataHandler.get_expected_target_repoids method is able to produce a correct
    map that maps target pesid to the best candidate pesid repository when one source pesid is mapped
    to multiple target pesids (every target pesid should have an entry in the returned map).
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2', 'pesid3'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga'),
            make_pesid_repo('pesid3', '8', 'pesid3-repoid-ga')
        ]
    )

    fail_description = 'Failed to get_expected_target_repoids when one source pesid is mapped to two target pesids.'
    handler = RepoMapDataHandler(repositories_mapping)
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-ga'])

    assert {'pesid2': repositories_mapping.repositories[1],
            'pesid3': repositories_mapping.repositories[2]} == target_repoids, fail_description


def test_get_expected_target_pesid_repos_unmapped_repository(monkeypatch):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the RepoMapDataHandler.get_expected_target_repoids method does not fail
    when there is a repository on the source system that is not mapped.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga')
        ])

    fail_description = 'Failed to get_expected_target_repoids when one of the source repoids is unmapped.'
    handler = RepoMapDataHandler(repositories_mapping)
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-ga', 'unmapped-repoid'])

    assert {'pesid2': repositories_mapping.repositories[1]} == target_repoids, fail_description


def test_get_expected_target_pesid_repos_repo_with_no_equivalent(monkeypatch, caplog):
    """
    Test for the RepoMapDataHandler.get_expected_target_repoids method.

    Verifies that the RepoMapDataHandler.get_expected_target_repoids method does not fail
    when there is a repository on the source system that does not have any equivalents.
    A warning should be produced when a situation like this occurs.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-eus', channel='eus'),
        ]
    )

    handler = RepoMapDataHandler(repositories_mapping)
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-ga'])

    fail_description = (
        'Failed get_expected_target_repoids with a source repository that does not have any target equivalent.')
    assert {'pesid2': None} == target_repoids, fail_description
    missing_target_equivalent_message = (
        'Cannot find any mapped target repository from the pesid2 family for the pesid1-repoid-ga repository.'
    )

    # A warning should be produced when a target equivalent was not found.
    warning_produced = False
    for record in caplog.records:
        if record.levelno == logging.WARNING and record.message == missing_target_equivalent_message:
            warning_produced = True
            break
    assert warning_produced, 'A warning should be produced when a repository has no equivalent.'


def test_get_default_repository_channels_simple(monkeypatch):
    """
    Test for the get_default_repository_channels function.

    Verifies that the function returns correct list of default channels on a source system
    where there is only one repository enabled from the pesid family in which are
    the default repositories searched in.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    repository_mapping = RepositoriesMapping(
        mapping=[],
        repositories=[make_pesid_repo('rhel7-base', '7', 'rhel7-repoid-ga', channel='ga')]
    )
    handler = RepoMapDataHandler(repository_mapping)

    assert ['ga'] == get_default_repository_channels(handler, ['rhel7-repoid-ga'])


def test_get_default_repository_channels_highest_priority_channel(monkeypatch):
    """
    Test for the get_default_repository_channels function.

    Verifies that the returned list contains the highest priority channel if there is a repository
    with the channel enabled on the source system.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    repository_mapping = RepositoriesMapping(
        mapping=[],
        repositories=[
            make_pesid_repo('rhel7-base', '7', 'rhel7-repoid-ga', channel='ga'),
            make_pesid_repo('rhel7-base', '7', 'rhel7-repoid-eus', channel='eus'),
        ]
    )
    handler = RepoMapDataHandler(repository_mapping)

    assert ['eus', 'ga'] == get_default_repository_channels(handler, ['rhel7-repoid-ga', 'rhel7-repoid-eus'])


def test_get_default_repository_channels_no_default_pesid_repo(monkeypatch):
    """
    Test for the get_default_repository_channels function.

    Verifies that the returned list contains some fallback channel even if no repository from the default
    pesid family in which are the channels searched is enabled.
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4'))
    repository_mapping = RepositoriesMapping(
        mapping=[],
        repositories=[
            make_pesid_repo('rhel7-base', '7', 'rhel7-repoid-ga', channel='ga'),
            make_pesid_repo('rhel7-base', '7', 'rhel7-repoid-eus', channel='eus'),
        ]
    )
    handler = RepoMapDataHandler(repository_mapping)

    assert ['ga'] == get_default_repository_channels(handler, ['some-repoid'])


def test_find_repository_equivalent_with_priority_channel(monkeypatch):
    """
    Tests whether the _find_repository_target_equivalent correctly respects the chosen preferred channel.
    """
    envars = {'LEAPP_DEVEL_TARGET_PRODUCT_TYPE': 'eus'}

    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4', envars=envars))
    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga', channel='ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-eus', channel='eus'),
        ]
    )

    handler = RepoMapDataHandler(repositories_mapping)
    handler.set_default_channels(['ga'])

    assert handler.prio_channel == 'eus'

    fail_description = '_find_repository_target_equivalent does not correctly respect preferred channel.'
    expected_target_equivalent = repositories_mapping.repositories[2]
    actual_target_equivalent = handler._find_repository_target_equivalent(repositories_mapping.repositories[0],
                                                                          'pesid2')
    assert expected_target_equivalent == actual_target_equivalent, fail_description


def test_get_expected_target_pesid_repos_with_priority_channel_set(monkeypatch):
    """
    Tests whether the get_expected_target_peid_repos correctly respects the chosen preferred channel.
    """

    envars = {'LEAPP_DEVEL_TARGET_PRODUCT_TYPE': 'eus'}

    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.4', envars=envars))

    repositories_mapping = RepositoriesMapping(
        mapping=[RepoMapEntry(source='pesid1', target=['pesid2', 'pesid3'])],
        repositories=[
            make_pesid_repo('pesid1', '7', 'pesid1-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-ga'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-eus', channel='eus'),
            make_pesid_repo('pesid2', '8', 'pesid2-repoid-aus', channel='aus'),
            make_pesid_repo('pesid3', '8', 'pesid3-repoid-ga')
        ]
    )

    handler = RepoMapDataHandler(repositories_mapping)
    # Set defaults to verify that the priority channel is not overwritten by defaults
    handler.set_default_channels(['aus', 'ga'])
    target_repoids = handler.get_expected_target_pesid_repos(['pesid1-repoid-ga'])

    fail_description = 'get_expected_target_peid_repos does not correctly respect preferred channel.'
    assert {'pesid2': repositories_mapping.repositories[2],
            'pesid3': repositories_mapping.repositories[4]} == target_repoids, fail_description


@pytest.mark.parametrize('rhui', ('', 'aws', 'aws-sap-e4s', 'azure', 'azure-sap-ha', 'azure-sap-apps'))
def test_multiple_repoids_in_repomapping(monkeypatch, rhui):
    """
    Tests whether a correct repository is selected when running on cloud with multiple repositories having the same ID.

    In such a case, the actor should use the cloud provider as a guide on which of the repositores should it pick.
    """

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.6'))

    mk_rhui_el7_pesid_repo = functools.partial(PESIDRepositoryEntry,
                                               pesid='rhel7-rhui',
                                               major_version='7',
                                               repoid='repoid7-rhui',
                                               repo_type='rpm',
                                               arch='x86_64',
                                               channel='ga')

    mk_rhui_el8_pesid_repo = functools.partial(PESIDRepositoryEntry,
                                               pesid='rhel8-rhui',
                                               major_version='8',
                                               repo_type='rpm',
                                               arch='x86_64',
                                               channel='ga')

    repomap = RepositoriesMapping(
        mapping=[RepoMapEntry(source='rhel7-rhui', target=['rhel8-rhui'])],
        repositories=[
            mk_rhui_el7_pesid_repo(rhui=''),
            mk_rhui_el7_pesid_repo(rhui='aws'),
            mk_rhui_el7_pesid_repo(rhui='azure'),
            mk_rhui_el8_pesid_repo(repoid='repoid8-rhui', rhui=''),
            mk_rhui_el8_pesid_repo(repoid='repoid8-rhui-aws', rhui='aws'),
            mk_rhui_el8_pesid_repo(repoid='repoid8-rhui-azure', rhui='azure'),
        ]
    )

    handler = RepoMapDataHandler(repomap, cloud_provider=rhui)
    target_repoids = handler.get_expected_target_pesid_repos(['repoid7-rhui'])

    assert len(target_repoids) == 1

    expected_suffixes = {
        '': '',
        'aws': '-aws',
        'aws-sap-e4s': '-aws',
        'azure': '-azure',
        'azure-sap-apps': '-azure',
        'azure-sap-ha': '-azure'
    }

    assert 'rhel8-rhui' in target_repoids
    assert target_repoids['rhel8-rhui'].repoid == 'repoid8-rhui{0}'.format(expected_suffixes[rhui])



import re


def _build_regex(pattern):
    regex = '^'
    part_beginning = 0
    while part_beginning < len(pattern):
        ix1 = pattern.find('*', part_beginning)
        ix2 = pattern.find('?', part_beginning)
        ix1 = len(pattern) if ix1 < 0 else ix1
        ix2 = len(pattern) if ix2 < 0 else ix2
        part_end = min(ix1, ix2)

        regex += re.escape(pattern[part_beginning:part_end])

        if part_end < len(pattern):
            if pattern[part_end] == '*':
                regex += '.*'
            else:
                regex += '.'

        part_beginning = part_end + 1

    regex += '$'
    return regex


def _pattern_matches(pattern, string):
    if pattern.lower() == 'all':
        return True
    regex = _build_regex(pattern)
    return re.match(regex, string, re.IGNORECASE) is not None


def _daemon_list_matches_daemon(daemon_list, daemon, recursion_depth):
    try:
        cur_list_end = daemon_list.index('except')
    except ValueError:
        cur_list_end = len(daemon_list)
    cur_list = daemon_list[:cur_list_end]
    matches_cur_list = False
    for item in cur_list:
        try:
            ix = item.index('@')
            # For simplicity, we ignore the host part. So we must make sure
            # that a daemon list containing a host-based pattern will always match
            # the daemon part of that host-based pattern (e.g. 'all except vsftpd@localhost
            # matches 'vsftpd'). See test_config_applies_to_daemon_with_host_except().
            if recursion_depth % 2 == 1:
                continue
            pattern = item[:ix]
        except ValueError:
            pattern = item
        if _pattern_matches(pattern, daemon):
            matches_cur_list = True
            break

    next_list = daemon_list[cur_list_end + 1:]
    if not next_list:
        matches_next_list = False
    else:
        matches_next_list = _daemon_list_matches_daemon(next_list, daemon, recursion_depth + 1)

    return matches_cur_list and not matches_next_list


def config_applies_to_daemon(facts, daemon):
    """
    Returns True if the specified tcp_wrappers configuration applies to the specified daemon.
    Otherwise returns False.

    This information is intended to be used in the Checks phase to check whether there is
    any tcp_wrappers configuration that the user needs to migrate manually and whether we
    should inhibit the upgrade, so that the upgraded system is not insecure.

    :param facts: A TcpWrappersFacts representation of the tcp_wrappers configuration
    :param daemon: The daemon name
    """
    for daemon_list in facts.daemon_lists:
        value = [item.lower() for item in daemon_list.value]
        if _daemon_list_matches_daemon(value, daemon, 0):
            return True
    return False



from leapp import reporting
from leapp.libraries.actor import checksystemarch
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.utils.report import is_inhibitor


def test_valid_architectures(monkeypatch):
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch=architecture.ARCH_ACCEPTED[0]))

    checksystemarch.check_architecture()

    assert reporting.create_report.called == 0


def test_invalid_architecture(monkeypatch):
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='invalid_architecture'))

    checksystemarch.check_architecture()
    assert reporting.create_report.called == 1
    assert 'Unsupported architecture' in reporting.create_report.report_fields['title']
    assert 'Upgrade process is only supported' in reporting.create_report.report_fields['summary']
    assert reporting.create_report.report_fields['severity'] == 'high'
    assert is_inhibitor(reporting.create_report.report_fields)



from multiprocessing import Manager

from leapp.models import SatelliteFacts, SatellitePostgresqlFacts
from leapp.snactor.fixture import current_actor_context


class MockedRun(object):
    def __init__(self):
        self._manager = Manager()
        self.commands = self._manager.list()

    def __call__(self, cmd, *args, **kwargs):
        self.commands.append(cmd)
        return {}


def test_run_installer(monkeypatch, current_actor_context):
    mocked_run = MockedRun()
    monkeypatch.setattr('leapp.libraries.stdlib.run', mocked_run)
    current_actor_context.feed(SatelliteFacts(has_foreman=True,
                                              postgresql=SatellitePostgresqlFacts(local_postgresql=False)))
    current_actor_context.run()
    assert mocked_run.commands
    assert len(mocked_run.commands) == 1
    assert mocked_run.commands[0] == ['foreman-installer', '--disable-system-checks']


def test_run_installer_without_katello(monkeypatch, current_actor_context):
    mocked_run = MockedRun()
    monkeypatch.setattr('leapp.libraries.stdlib.run', mocked_run)
    current_actor_context.feed(SatelliteFacts(has_foreman=True, has_katello_installer=False,
                                              postgresql=SatellitePostgresqlFacts(local_postgresql=False)))
    current_actor_context.run()
    assert mocked_run.commands
    assert len(mocked_run.commands) == 1
    assert mocked_run.commands[0] == ['foreman-installer']


def test_run_reindexdb(monkeypatch, current_actor_context):
    mocked_run = MockedRun()
    monkeypatch.setattr('leapp.libraries.stdlib.run', mocked_run)
    current_actor_context.feed(SatelliteFacts(has_foreman=True,
                                              postgresql=SatellitePostgresqlFacts(local_postgresql=True)))
    current_actor_context.run()
    assert mocked_run.commands
    assert len(mocked_run.commands) == 4
    assert mocked_run.commands[0] == ['sed', '-i', '/data_directory/d', '/var/lib/pgsql/data/postgresql.conf']
    assert mocked_run.commands[1] == ['systemctl', 'start', 'postgresql']
    assert mocked_run.commands[2] == ['runuser', '-u', 'postgres', '--', 'reindexdb', '-a']
    assert mocked_run.commands[3] == ['foreman-installer', '--disable-system-checks']



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import version
from leapp.libraries.stdlib import api
from leapp.models import (
    SystemdServicesInfoSource,
    SystemdServicesInfoTarget,
    SystemdServicesPresetInfoSource,
    SystemdServicesPresetInfoTarget,
    SystemdServicesTasks
)

FMT_LIST_SEPARATOR = "\n    - "


def _get_desired_service_state(state_source, preset_source, preset_target):
    """
    Get the desired service state on the target system

    :param state_source: State on the source system
    :param preset_source: Preset on the source system
    :param preset_target: Preset on the target system
    :return: The desired state on the target system
    """

    if state_source in ("disabled", "enabled-runtime"):
        if preset_source == "disable":
            return preset_target + "d"  # use the default from target

    return state_source


def _get_desired_states(
    services_source, presets_source, services_target, presets_target
):
    "Get the states that services should be in on the target system"
    desired_states = {}

    for service in services_target:
        state_source = services_source.get(service.name)
        preset_target = _get_service_preset(service.name, presets_target)
        preset_source = _get_service_preset(service.name, presets_source)

        desired_state = _get_desired_service_state(
            state_source, preset_source, preset_target
        )
        desired_states[service.name] = desired_state

    return desired_states


def _get_service_task(service_name, desired_state, state_target, tasks):
    """
    Get the task to set the desired state of the service on the target system

    :param service_name: Then name of the service
    :param desired_state: The state the service should set to
    :param state_target: State on the target system
    :param tasks: The tasks to append the task to
    """
    if desired_state == state_target:
        return

    if desired_state == "enabled":
        tasks.to_enable.append(service_name)
    if desired_state == "disabled":
        tasks.to_disable.append(service_name)


def _get_service_preset(service_name, presets):
    preset = presets.get(service_name)
    if not preset:
        # shouldn't really happen as there is usually a `disable *` glob as
        # the last statement in the presets
        api.current_logger().debug(
            'No presets found for service "{}", assuming "disable"'.format(service_name)
        )
        return "disable"
    return preset


def _filter_ignored_services(services_source):
    """
    Filter out services that should be ignored i.e. not handled
    """
    to_ignore = []
    if int(version.get_source_major_version()) >= 8:
        to_ignore.extend([
            "libvirtd.service",
            "virtqemud.service",
            "virtinterfaced.service",
            "virtnetworkd.service",
            "virtnodedevd.service",
            "virtnwfilterd.service",
            "virtsecretd.service",
            "virtstoraged.service",
            "virtproxyd.service",
            "virtlockd.service",
            "virtlogd.service",
            "libvirt-guests.service",
        ])

    for s in to_ignore:
        # It's sufficient to remove just from the source system services,
        # because if a service is not present on the source system it's not handled either way
        if services_source.pop(s, None):
            api.current_logger().debug("Ignored service {} found on the source system".format(s))


def _filter_irrelevant_services(services_source, services_target):
    """
    Filter out irrelevant services

    Irrelevant services are those that cannot be enabled/disabled,
    those that do not exist on the source system and those in masked-runtime state.

    :return: Target system services without the irrelevant ones.
    :rtype: list
    """
    filtered = []
    for service in services_target:
        if service.state not in ("enabled", "disabled", "enabled-runtime"):
            # Enabling/disabling of services is only relevant to these states
            continue

        state_source = services_source.get(service.name)
        if not state_source:
            # The service doesn't exist on the source system
            continue

        if state_source == "masked-runtime":
            # TODO(mmatuska): It's not possible to get the persistent
            # (non-runtime) state of a service with `systemctl`. One solution
            # might be to check symlinks
            api.current_logger().debug(
                'Skipping service in "masked-runtime" state: {}'.format(service.name)
            )
            continue

        filtered.append(service)

    return filtered


def _get_required_tasks(services_target, desired_states):
    """
    Get the required tasks to set the services on the target system to their desired state

    :return: The tasks required to be executed
    :rtype: SystemdServicesTasks
    """
    tasks = SystemdServicesTasks()

    for service in services_target:
        desired_state = desired_states[service.name]
        _get_service_task(service.name, desired_state, service.state, tasks)

    return tasks


def _report_kept_enabled(tasks):
    summary = (
        "Systemd services which were enabled on the system before the upgrade"
        " were kept enabled after the upgrade. "
    )
    if tasks:
        summary += (
            "The following services were originally disabled by preset on the"
            " upgraded system and Leapp attempted to enable them:{}{}"
        ).format(FMT_LIST_SEPARATOR, FMT_LIST_SEPARATOR.join(sorted(tasks.to_enable)))
        # TODO(mmatuska): When post-upgrade reports are implemented in
        # `setsystemdservicesstates actor, add a note here to check the reports
        # if the enabling failed

    reporting.create_report(
        [
            reporting.Title("Previously enabled systemd services were kept enabled"),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.INFO),
            reporting.Groups([reporting.Groups.POST]),
        ]
    )


def _get_newly_enabled(services_source, desired_states):
    newly_enabled = []
    for service, state in desired_states.items():
        state_source = services_source[service]
        if state_source == "disabled" and state == "enabled":
            newly_enabled.append(service)

    return newly_enabled


def _report_newly_enabled(newly_enabled):
    summary = (
        "The following services were disabled before the upgrade and were set"
        "to enabled by a systemd preset after the upgrade:{}{}.".format(
            FMT_LIST_SEPARATOR, FMT_LIST_SEPARATOR.join(sorted(newly_enabled))
        )
    )

    reporting.create_report(
        [
            reporting.Title("Some systemd services were newly enabled"),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.INFO),
            reporting.Groups([reporting.Groups.POST]),
        ]
    )


def _expect_message(model):
    """
    Get the expected message or throw an error
    """
    message = next(api.consume(model), None)
    if not message:
        raise StopActorExecutionError(
            "Expected {} message, but didn't get any".format(model.__name__)
        )
    return message


def process():
    services_source = _expect_message(SystemdServicesInfoSource).service_files
    services_target = _expect_message(SystemdServicesInfoTarget).service_files
    presets_source = _expect_message(SystemdServicesPresetInfoSource).presets
    presets_target = _expect_message(SystemdServicesPresetInfoTarget).presets

    services_source = {p.name: p.state for p in services_source}
    presets_source = {p.service: p.state for p in presets_source}
    presets_target = {p.service: p.state for p in presets_target}

    _filter_ignored_services(services_source)
    services_target = _filter_irrelevant_services(services_source, services_target)

    desired_states = _get_desired_states(
        services_source, presets_source, services_target, presets_target
    )
    tasks = _get_required_tasks(services_target, desired_states)

    api.produce(tasks)
    _report_kept_enabled(tasks)

    newly_enabled = _get_newly_enabled(services_source, desired_states)
    _report_newly_enabled(newly_enabled)



from leapp.actors import Actor
from leapp.libraries.actor import grub2mkconfigonppc64
from leapp.models import DefaultGrubInfo, FirmwareFacts
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class Grub2MkconfigOnPpc64(Actor):
    """
    Regenerate Grub config after RHEL 8 to RHEL 9 rpm transaction

    Actor runs 'grub2-mkconfig' to regenerate Grub config after RHEL 8 to RHEL 9
    rpm transaction in case system is running on ppc64le, GRUB_ENABLE_BLSCFG=true
    and '/boot/grub2/grub.cfg' config file does not yet work with BLS entries.
    For the system to successfully boot into el9 kernel, it is essential that the
    grub config file is abvle to work with the BLS entries.

    IMPORTANT NOTE: The fix is applied only for virtualized ppc64le systems as we got
    unexpected behavior on bare metal ppc64le systems which needs to be
    investigated first.

    """

    name = 'grub2mkconfig_on_ppc64'
    consumes = (DefaultGrubInfo, FirmwareFacts)
    produces = ()
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        grub2mkconfigonppc64.process()



import contextlib
import os
import shutil

import pytest

from leapp.libraries.actor import quaggatofrr
from leapp.libraries.common.testutils import CurrentActorMocked

ACTIVE_DAEMONS = ['bgpd', 'ospfd', 'zebra']
CUR_DIR = os.path.dirname(os.path.abspath(__file__))
FROM_DIR = '/tmp/from_dir/'
TO_DIR = '/tmp/to_dir/'
CONFIG_DATA = {
    'bgpd': '--daemon -A 10.10.100.1',
    'isisd': '--daemon -A ::1',
    'ospf6d': '-A ::1',
    'ospfd': '-A 127.0.0.1',
    'ripd': '-A 127.0.0.1',
    'ripngd': '-A ::1',
    'zebra': '-s 90000000 --daemon -A 127.0.0.1'
}


@contextlib.contextmanager
def _create_mock_files():
    try:
        os.mkdir(FROM_DIR)
        os.mkdir(TO_DIR)

        for num in range(1, 10):
            full_path = "{}test_file_{}".format(FROM_DIR, num)
            with open(full_path, 'w') as fp:
                fp.write("test_file_{}".format(num))
        yield
    finally:
        shutil.rmtree(FROM_DIR)
        shutil.rmtree(TO_DIR)


class MockedFilePointer(object):
    def __init__(self, orig_open, fname, mode='r'):
        self._orig_open = orig_open
        self.fname = fname
        self.mode = mode
        # we want always read only..
        self._fp = self._orig_open(fname, 'r')
        self._read = None
        self.written = None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def close(self):
        if self._fp:
            self._fp.close()
            self._fp = None

    def read(self):
        self._read = self._fp.read()
        return self._read

    def write(self, data):
        if not self.written:
            self.written = data
        else:
            self.written += data


class MockedOpen(object):
    """
    This is mock for the open function. When called it creates
    the MockedFilePointer object.
    """

    def __init__(self):
        # currently we want to actually open the real files, we need
        # to mock other stuff related to file pointers / file objects
        self._orig_open = open
        self._open_called = []

    def __call__(self, fname, mode='r'):
        opened = MockedFilePointer(self._orig_open, fname, mode)
        self._open_called.append(opened)
        return opened

    def get_mocked_pointers(self, fname, mode=None):
        """
        Get list of MockedFilePointer objects with the specified fname.

        if the mode is set (expected 'r', 'rw', 'w' ..) discovered files are
        additionally filtered to match the same mode (same string).
        """
        fnames = [i for i in self._open_called if i.fname == fname]
        return fnames if not mode else [i for i in fnames if i.mode == mode]


def test_copy_config_files():
    with _create_mock_files():
        quaggatofrr._copy_config_files(FROM_DIR, TO_DIR)
        conf_files = os.listdir(TO_DIR)
        for file_name in conf_files:
            full_path = os.path.join(TO_DIR, file_name)
            assert os.path.isfile(full_path)


def test_get_config_data():
    conf_data = quaggatofrr._get_config_data(
        os.path.join(CUR_DIR, 'files/quagga')
    )

    assert 'babels' not in conf_data
    assert conf_data['bgpd'] == CONFIG_DATA['bgpd']
    assert conf_data['isisd'] == CONFIG_DATA['isisd']
    assert conf_data['ospf6d'] == CONFIG_DATA['ospf6d']
    assert conf_data['ospfd'] == CONFIG_DATA['ospfd']
    assert conf_data['ripd'] == CONFIG_DATA['ripd']
    assert conf_data['ripngd'] == CONFIG_DATA['ripngd']
    assert conf_data['zebra'] == CONFIG_DATA['zebra']


def test_edit_new_config():
    # writing the data to the new config file
    data = quaggatofrr._edit_new_config(
        os.path.join(CUR_DIR, 'files/daemons'),
        ACTIVE_DAEMONS,
        CONFIG_DATA
    )

    assert 'zebra=yes' in data
    assert 'bgpd=yes' in data
    assert 'ospfd=yes' in data
    assert 'zebra_options=("-s 90000000 --daemon -A 127.0.0.1")' in data
    assert 'bgpd_options=("--daemon -A 10.10.100.1")' in data
    assert 'ospfd_options=("-A 127.0.0.1")' in data
    assert 'ospf6d_options=("-A ::1")' in data
    assert 'ripd_options=("-A 127.0.0.1")' in data
    assert 'ripngd_options=("-A ::1")' in data
    assert 'isisd_options=("--daemon -A ::1")' in data


@pytest.mark.parametrize('dst_ver', ['8.4', '8.5'])
def test_fix_commands(monkeypatch, dst_ver):
    monkeypatch.setattr(quaggatofrr, "BGPD_CONF_FILE", os.path.join(CUR_DIR, 'files/bgpd.conf'))
    monkeypatch.setattr(quaggatofrr.api, 'current_actor', CurrentActorMocked(dst_ver=dst_ver))
    monkeypatch.setattr(quaggatofrr, "open", MockedOpen(), False)
    quaggatofrr._fix_commands()

    fp_list = quaggatofrr.open.get_mocked_pointers(quaggatofrr.BGPD_CONF_FILE, "w")
    assert len(fp_list) == 1
    assert 'bgp extcommunity-list' in fp_list[0].written


def test_fix_commands_not_applied(monkeypatch):
    is_file_called = False

    def mocked_is_file(dummy):
        is_file_called = True
        return is_file_called

    monkeypatch.setattr(quaggatofrr.api, 'current_actor', CurrentActorMocked(dst_ver='8.3'))
    monkeypatch.setattr(os.path, 'isfile', mocked_is_file)
    monkeypatch.setattr(quaggatofrr, "open", MockedOpen(), False)
    quaggatofrr._fix_commands()
    assert not is_file_called



import pytest

from leapp.models import DistributionSignedRPM, Report, RPM


def _generate_rpm_with_name(name):
    return RPM(name=name,
               version='0.1',
               release='1.sm01',
               epoch='1',
               pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51',
               packager='Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>',
               arch='noarch')


@pytest.mark.parametrize('unsupported_versions', [
    ([]),          # No unsupported versions
    ([2.1]),       # Single unsupported version
    ([3.0]),       # Other unsupported version
    ([2.1, 3.0]),  # Multiple unsupported versions
])
def test_actor_execution(monkeypatch, current_actor_context, unsupported_versions):
    """
    Install one or more dotnet-runtime packages for unsupported versions
    and verify a report is generated.
    """

    # Couple of random packages
    rpms = [_generate_rpm_with_name('sed'),
            _generate_rpm_with_name('htop')]

    # dotnet-runtime-{version} packages
    for version in unsupported_versions:
        rpms += [_generate_rpm_with_name(f'dotnet-runtime-{version}')]

    # Executed actor fed with fake RPMs
    current_actor_context.feed(DistributionSignedRPM(items=rpms))
    current_actor_context.run()

    if unsupported_versions:
        assert current_actor_context.consume(Report)
    else:
        assert not current_actor_context.consume(Report)



import pytest

from leapp.libraries.actor.add_provider import (
    _add_lines,
    _append,
    _modify_file,
    _replace,
    APPEND_STRING,
    LEAPP_COMMENT,
    NotFoundException
)

testdata = (
    ([], 'one', ['one\n']),
    (['first'], 'one', ['first', 'one\n']),
    (['first'], 'one\ntwo', ['first', 'one\n', 'two\n']),
)


@pytest.mark.parametrize('lines,add,expected', testdata)
def test_add_lines(lines, add, expected):
    r = _add_lines(lines, add)
    assert r == expected


testdata = (
    ([], "search", "replace", None, False, None),
    ([], "search", "replace", "comment", False, None),
    ([], r"\s*", "replace", None, False, None),
    ([], r"\s*", "replace", "comment", False, None),
    (["text"], "text", "replace", None, False, ["replace\n"]),
    (["text"], "text", "replace", "comment", False, ["comment", "replace\n"]),
    (["text"], "text", "replace", "comment", True, ["comment", "# text", "replace\n"]),
    (["text    "], "text", "replace", "comment", True, ["comment", "# text    ", "replace\n"]),
    (["    text"], "text", "replace", "comment", True, ["comment", "#     text", "replace\n"]),
    (["text    text"], r"text\s*text", "replace", "comment", True, ["comment", "# text    text", "replace\n"]),
    (["first", "text", "last"], "text", "replace", "comment", True,
     ["first", "comment", "# text", "replace\n", "last"]),
)


@pytest.mark.parametrize('lines,search,replace,comment,backup,expected', testdata)
def test_replace(lines, search, replace, comment, backup, expected):
    try:
        r = _replace(lines, search, replace, comment, backup)
        if expected:
            assert r == expected
        else:
            assert False
    except NotFoundException:
        assert not expected


testdata = (
    ([], 'one', 'comment', ['comment', 'one\n']),
    (['first'], 'one', 'comment', ['first', 'comment', 'one\n']),
    (['first'], 'one\ntwo', 'comment', ['first', 'comment', 'one\n', 'two\n']),
)


@pytest.mark.parametrize('lines,add,comment,expected', testdata)
def test_append(lines, add, comment, expected):
    r = _append(lines, add, comment)
    assert r == expected


class MockFile(object):
    def __init__(self, content=None):
        self.content = content
        self.error = False

    def readlines(self):
        return self.content.splitlines(True)

    def seek(self, n):
        self.content = ''

    def write(self, content):
        self.content = content


testdata = (
    ('', ''),
    ('openssl_conf=default_modules\n',
     '{}# openssl_conf=default_modules\nopenssl_conf = openssl_init\n'.format(LEAPP_COMMENT)),
    ('openssl_conf = default_modules\n',
     '{}# openssl_conf = default_modules\nopenssl_conf = openssl_init\n'.format(LEAPP_COMMENT)),
    ('openssl_conf  =  default_modules\n',
     '{}# openssl_conf  =  default_modules\nopenssl_conf = openssl_init\n'.format(LEAPP_COMMENT)),
    ('  openssl_conf = default_modules  \n',
     '{}#   openssl_conf = default_modules  \nopenssl_conf = openssl_init\n'.format(LEAPP_COMMENT)),
    ('[default_modules]\n',
     '{}# [default_modules]\n[openssl_init]\nproviders = provider_sect\n'.format(LEAPP_COMMENT)),
    ('[  default_modules  ]\n',
     '{}# [  default_modules  ]\n[openssl_init]\nproviders = provider_sect\n'.format(LEAPP_COMMENT)),
    ('  [ default_modules ] \n',
     '{}#   [ default_modules ] \n[openssl_init]\nproviders = provider_sect\n'.format(LEAPP_COMMENT)),
    ('openssl_conf=default_modules\n[default_modules]\n',
     '{c}# openssl_conf=default_modules\nopenssl_conf = openssl_init\n'
     '{c}# [default_modules]\n[openssl_init]\nproviders = provider_sect\n'.format(c=LEAPP_COMMENT)),
)


@pytest.mark.parametrize('file_content,expected', testdata)
def test_modify_file(file_content, expected):
    f = MockFile(file_content)

    # Test separate replaces and do not fail if pattern is not found
    _modify_file(f, False)

    assert f.content == "{}{}{}\n".format(expected, LEAPP_COMMENT, APPEND_STRING)



try:
    from rpm import labelCompare
except ImportError:
    # this can happen only on non-rpm based systems or with Python3 on RHEL 7
    # based OSs, as the rpm python module is available here just for Python2
    # - and vice versa on F31+
    # This will not happen in real run, just in case of unit-tests..
    def labelCompare(*args):
        raise NotImplementedError(
            "The labelCompare function is not implemented for the python"
            " you are using."
        )

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture, utils
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, KernelInfo


def get_all_pkgs_with_name(pkg_name):
    """
    Get all installed packages of the given name signed by Red Hat.
    """
    rpms = next(api.consume(DistributionSignedRPM), DistributionSignedRPM()).items
    return [pkg for pkg in rpms if pkg.name == pkg_name]


def get_EVR(pkg):
    """
    Return 3-tuple EVR (_epoch_, version, release) of the given RPM.

    Epoch is always set as an empty string as in case of kernel epoch is not
    expected to be set - ever.
    """
    return ('', pkg.version, pkg.release)


def get_newest_evr(pkgs):
    """
    Return the 3-tuple (EVR) of the newest package from the given list.

    Return None if the given list is empty. It's expected that all given
    packages have same name.
    """
    if not pkgs:
        return None

    newest_evr = get_EVR(pkgs[0])
    for pkg in pkgs:
        evr = get_EVR(pkg)
        if labelCompare(newest_evr, evr) < 0:
            newest_evr = evr

    return newest_evr


def process():
    kernel_info = utils._require_exactly_one_message_of_type(KernelInfo)
    pkgs = get_all_pkgs_with_name(kernel_info.pkg.name)

    if not pkgs:
        # Hypothetical, user is not allowed to install any kernel that is not signed by RH
        # In case we would like to be cautious, we could check whether there are no other
        # kernels installed as well.
        api.current_logger().error('Cannot find any installed kernel signed by Red Hat.')
        raise StopActorExecutionError('Cannot find any installed kernel signed by Red Hat.')

    if len(pkgs) > 1 and architecture.matches_architecture(architecture.ARCH_S390X):
        # It's temporary solution, so no need to try automate everything.
        title = 'Multiple kernels installed'
        summary = ('The upgrade process does not handle well the case when multiple kernels'
                   ' are installed on s390x. There is a severe risk of the bootloader configuration'
                   ' getting corrupted during the upgrade.')
        remediation = ('Boot into the most up-to-date kernel and remove all older'
                       ' kernels installed on the machine before running Leapp again.')
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.BOOT]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Remediation(hint=remediation),
            reporting.RelatedResource('package', 'kernel')
        ])

    current_kernel_evr = get_EVR(kernel_info.pkg)
    newest_kernel_evr = get_newest_evr(pkgs)

    api.current_logger().debug('Current kernel EVR: {}'.format(current_kernel_evr))
    api.current_logger().debug('Newest kernel EVR: {}'.format(newest_kernel_evr))

    if current_kernel_evr != newest_kernel_evr:
        title = 'Newest installed kernel not in use'
        summary = ('To ensure a stable upgrade, the machine needs to be'
                   ' booted into the latest installed kernel.')
        remediation = ('Boot into the most up-to-date kernel installed'
                       ' on the machine before running Leapp again.')
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.BOOT]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Remediation(hint=remediation),
            reporting.RelatedResource('package', 'kernel')
        ])



import os

from leapp.libraries.stdlib import api, CalledProcessError, run

CUSTOM_DNF_CONF_PATH = "/etc/leapp/files/dnf.conf"


def process():
    if os.path.exists(CUSTOM_DNF_CONF_PATH):
        try:
            run(["mv", CUSTOM_DNF_CONF_PATH, "/etc/dnf/dnf.conf"])
        except (CalledProcessError, OSError) as e:
            api.current_logger().debug(
                "Failed to move /etc/leapp/files/dnf.conf to /etc/dnf/dnf.conf: {}".format(e)
            )



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class LDConfigFile(Model):
    """
    Represents a config file related to dynamic linker configuration
    """
    topic = SystemFactsTopic

    path = fields.String()
    """ Absolute path to the configuration file """

    modified = fields.Boolean()
    """ If True the file is considered custom and will trigger a report """


class MainLDConfigFile(LDConfigFile):
    """
    Represents the main configuration file of the dynamic linker /etc/ld.so.conf
    """
    topic = SystemFactsTopic

    modified_lines = fields.List(fields.String(), default=[])
    """ Lines that are considered custom, generally those that are not includes of other configs """


class DynamicLinkerConfiguration(Model):
    """
    Facts about configuration of dynamic linker
    """
    topic = SystemFactsTopic

    main_config = fields.Model(MainLDConfigFile)
    """ The main configuration file of dynamic linker (/etc/ld.so.conf) """

    included_configs = fields.List(fields.Model(LDConfigFile))
    """ All the configs that are included by the main configuration file """

    used_variables = fields.List(fields.String(), default=[])
    """ Environment variables that are currently used to modify dynamic linker configuration """



import re

from leapp.libraries.stdlib import api, run
from leapp.models import (
    ActiveKernelModulesFacts,
    DetectedDeviceOrDriver,
    DeviceDriverDeprecationData,
    PCIDevice,
    PCIDevices
)

# Regex to capture Vendor, Device and SVendor and SDevice values
PCI_ID_REG = re.compile(r"(?<=Vendor:\t|Device:\t)\w+")


# TODO this could be solved more efficiently and error prune via python re
#   and groupdict
def parse_pci_device(textual_block, numeric_block):
    """ Parse one block from lspci output describing one PCI device """
    device = {
        'Slot': '',
        'Class': '',
        'Vendor': '',
        'Device': '',
        'SVendor': '',
        'SDevice': '',
        'PhySlot': '',
        'Rev': '',
        'ProgIf': '',
        'Driver': '',
        'Module': [],
        'NUMANode': ''
    }
    for line in textual_block.splitlines():
        key, value = line.split(':\t')

        if key in device:
            if isinstance(device[key], list):
                device[key].append(value)
            else:
                if device[key]:
                    api.current_logger().debug(
                        'Unexpected duplicate key - {k}: {v} (current value: {vcurr}), ignoring'.format(
                            k=key, v=value, vcurr=device[key]))
                else:
                    device[key] = value
        else:
            api.current_logger().debug('Unrecognized key - {k}: {v}, ignoring'.format(k=key, v=value))

    return PCIDevice(
        slot=device['Slot'],
        dev_cls=device['Class'],
        vendor=device['Vendor'],
        name=device['Device'],
        subsystem_vendor=device['SVendor'],
        subsystem_name=device['SDevice'],
        physical_slot=device['PhySlot'],
        rev=device['Rev'],
        progif=device['ProgIf'],
        driver=device['Driver'],
        modules=device['Module'],
        numa_node=device['NUMANode'],
        pci_id=":".join(PCI_ID_REG.findall(numeric_block))
    )


def parse_pci_devices(pci_textual, pci_numeric):
    """ Parse lspci output and return a list of PCI devices """
    return [
        parse_pci_device(*block) for block
        in zip(
            pci_textual.split('\n\n')[:-1],
            pci_numeric.split('\n\n')[:-1]
        )
    ]


def produce_detected_devices(devices):
    prefix_re = re.compile('0x')
    entry_lookup = {
        prefix_re.sub('', entry.device_id): entry
        for message in api.consume(DeviceDriverDeprecationData) for entry in message.entries
    }
    api.produce(*[
        DetectedDeviceOrDriver(**entry_lookup[device.pci_id].dump())
        for device in devices
        if device.pci_id in entry_lookup
    ])


def produce_detected_drivers(devices):
    active_modules = {
        module.file_name
        for message in api.consume(ActiveKernelModulesFacts) for module in message.kernel_modules
    }

    # Create a lookup by driver_name and filter out the kernel that are active
    entry_lookup = {
        entry.driver_name: entry
        for message in api.consume(DeviceDriverDeprecationData) for entry in message.entries
        if not entry.device_id and entry.driver_name and entry.driver_name not in active_modules
    }

    drivers = {device.driver for device in devices if device.driver in entry_lookup}
    api.produce(*[
        DetectedDeviceOrDriver(**entry_lookup[driver].dump())
        for driver in drivers
    ])


def produce_pci_devices(producer, devices):
    """ Produce a Leapp message with all PCI devices """
    producer(PCIDevices(devices=devices))


def scan_pci_devices(producer):
    """ Scan system PCI Devices """
    pci_textual = run(['lspci', '-vmmk'], checked=False)['stdout']
    pci_numeric = run(['lspci', '-vmmkn'], checked=False)['stdout']
    devices = parse_pci_devices(pci_textual, pci_numeric)
    produce_detected_devices(devices)
    produce_detected_drivers(devices)
    produce_pci_devices(producer, devices)



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.config import is_verbose
from leapp.models import InstalledUnsignedRPM

COMMON_REPORT_TAGS = [reporting.Groups.SANITY]


def generate_report(packages):
    """ Generate a report if there are unsigned packages installed on the system """
    if not packages:
        return
    unsigned_packages_new_line = '\n'.join(['- ' + p for p in packages])
    title = 'Packages not signed by Red Hat found on the system'
    summary = ('The following packages have not been signed by Red Hat'
               ' and may be removed during the upgrade process in case Red Hat-signed'
               ' packages to be removed during the upgrade depend on them:\n{}'
               .format(unsigned_packages_new_line))
    reporting.create_report([
        reporting.Title(title),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups(COMMON_REPORT_TAGS)
    ])

    if is_verbose():
        api.show_message(summary)


def get_unsigned_packages():
    """ Get list of unsigned packages installed in the system """
    rpm_messages = api.consume(InstalledUnsignedRPM)
    data = next(rpm_messages, InstalledUnsignedRPM())
    if list(rpm_messages):
        api.current_logger().warning('Unexpectedly received more than one InstalledUnsignedRPM message.')
    unsigned_packages = set()
    unsigned_packages.update([pkg.name for pkg in data.items])
    unsigned_packages = list(unsigned_packages)
    unsigned_packages.sort()
    return unsigned_packages


def check_unsigned_packages():
    """ Check and generate reports if system contains unsigned installed packages"""
    generate_report(get_unsigned_packages())



from leapp.actors import Actor
from leapp.libraries.actor import scanblacklistca
from leapp.models import BlackListCA, BlackListError
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanBlackListCA(Actor):
    """
    Scan the file system for distrusted CA's in the blacklist directory.

    The will be moved to the corresponding blocklist directory as the blacklist
    directory is deprecated in RHEL-9
    """

    name = 'scanblacklistca'
    consumes = ()
    produces = (BlackListCA, BlackListError)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scanblacklistca.process()



import textwrap

from six import StringIO

from leapp.libraries.actor.sssdfacts import SSSDFactsLibrary
from leapp.libraries.common import utils
from leapp.models import SSSDConfig, SSSDDomainConfig


def get_config(content):
    parser = utils.parse_config(StringIO(textwrap.dedent(content)))
    return parser


def test_empty_config():
    config = utils.parse_config()
    facts = SSSDFactsLibrary(config).process()

    assert not facts.domains


def test_local_domain():
    config = get_config("""
    [domain/local]
    id_provider = local
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "local"
    assert len(facts.domains[0].options) == 1
    assert "local_provider" in facts.domains[0].options


def test_groups_chain():
    config = get_config("""
    [domain/ldap]
    id_provider = ldap
    ldap_groups_use_matching_rule_in_chain = True

    # Set this option to avoid reporting its change
    ldap_sudo_include_regexp = True
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "ldap"
    assert len(facts.domains[0].options) == 1
    assert "groups_chain" in facts.domains[0].options


def test_initgroups_chain():
    config = get_config("""
    [domain/ldap]
    id_provider = ldap
    ldap_initgroups_use_matching_rule_in_chain = True

    # Set this option to avoid reporting its change
    ldap_sudo_include_regexp = True
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "ldap"
    assert len(facts.domains[0].options) == 1
    assert "initgroups_chain" in facts.domains[0].options


def test_sudo_regexp__ldap():
    config = get_config("""
    [domain/ldap]
    id_provider = ldap
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "ldap"
    assert len(facts.domains[0].options) == 1
    assert "sudo_regexp" in facts.domains[0].options


def test_sudo_regexp__ad():
    config = get_config("""
    [domain/ad]
    id_provider = ad
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "ad"
    assert len(facts.domains[0].options) == 1
    assert "sudo_regexp" in facts.domains[0].options


def test_sudo_regexp__ipa():
    config = get_config("""
    [domain/ipa]
    id_provider = ipa
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 1
    assert facts.domains[0].name == "ipa"
    assert not facts.domains[0].options


def test_complex():
    config = get_config("""
    [sssd]
    user = root
    services = nss, pam, sudo

    [domain/local]
    id_provider = local

    [domain/ldap]
    id_provider = ldap
    ldap_groups_use_matching_rule_in_chain = True

    [domain/ipa]
    id_provider = ipa
    """)

    facts = SSSDFactsLibrary(config).process()

    assert len(facts.domains) == 3

    assert facts.domains[0].name == "local"
    assert len(facts.domains[0].options) == 1
    assert "local_provider" in facts.domains[0].options

    assert facts.domains[1].name == "ldap"
    assert len(facts.domains[1].options) == 2
    assert "groups_chain" in facts.domains[1].options
    assert "sudo_regexp" in facts.domains[1].options

    assert facts.domains[2].name == "ipa"
    assert not facts.domains[2].options


def test_get_domain_section():
    config = utils.parse_config()
    library = SSSDFactsLibrary(config)

    assert library.get_domain_section("ldap") == "domain/ldap"


def test_get_provider__none():
    config = get_config("""
    [domain/local]
    """)
    library = SSSDFactsLibrary(config)

    assert library.get_provider("local", "id_provider") is None


def test_get_provider__without_fallback():
    config = get_config("""
    [domain/local]
    id_provider = local
    """)
    library = SSSDFactsLibrary(config)

    assert library.get_provider("local", "id_provider") == "local"


def test_get_provider__with_fallback():
    config = get_config("""
    [domain/ldap]
    id_provider = ldap
    """)
    library = SSSDFactsLibrary(config)

    assert library.get_provider("ldap", "sudo_provider", ["id_provider"]) == "ldap"



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import StorageInfo


def inhibit_upgrade_due_non_persistent_mount(mountpoint):
    summary = (
        'Leapp detected that the {0} mountpoint is mounted, with no corresponding entry in /etc/fstab. '
        'The upgrade process cannot continue with {0} mounted in non-persistent fashion, '
        'as Leapp needs this mount to be available after a reboot.'
    )

    hint = (
        'Add {0} mount entry to /etc/fstab'
    )

    reporting.create_report([
        reporting.Title(
            'Detected partitions mounted in a non-persistent fashion, preventing a successful in-place upgrade.'
        ),
        reporting.Summary(summary.format(mountpoint)),
        reporting.Remediation(hint=hint.format(mountpoint)),
        reporting.RelatedResource('file', '/etc/fstab'),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.FILESYSTEM]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
    ])


def check_mount_is_persistent(storage_info, mountpoint):
    """Check if mountpoint is mounted in persistent fashion"""

    mount_entry_exists = any(me.mount == mountpoint for me in storage_info.mount)
    fstab_entry_exists = any(fe.fs_file == mountpoint for fe in storage_info.fstab)

    if mount_entry_exists and not fstab_entry_exists:
        inhibit_upgrade_due_non_persistent_mount(mountpoint)


def check_persistent_mounts():
    storage_info = next(api.consume(StorageInfo), None)
    if storage_info:
        check_mount_is_persistent(storage_info, '/var/lib/leapp')



from __future__ import division

from os import statvfs

from leapp import reporting

MIN_AVAIL_BYTES_FOR_BOOT = 100 * 2**20  # 100 MiB


def check_avail_space_on_boot(boot_avail_space_getter):
    avail_bytes = boot_avail_space_getter()
    if is_additional_space_required(avail_bytes):
        inhibit_upgrade(avail_bytes)


def get_avail_bytes_on_boot():
    boot_stat = statvfs('/boot')
    return boot_stat.f_frsize * boot_stat.f_bavail


def is_additional_space_required(avail_bytes):
    return avail_bytes < MIN_AVAIL_BYTES_FOR_BOOT


def inhibit_upgrade(avail_bytes):
    additional_mib_needed = (MIN_AVAIL_BYTES_FOR_BOOT - avail_bytes) / 2**20
    # we use "reporting.report_generic" to allow mocking in the tests
    # WIP ^^ check if this still applies
    reporting.create_report([
        reporting.Title('Not enough space on /boot'),
        reporting.Summary(
            '/boot needs additional {0} MiB to be able to accommodate the upgrade initramfs and new kernel.'.format(
             additional_mib_needed)
        ),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.FILESYSTEM]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.RelatedResource('directory', '/boot')
    ])



import os

import pytest

from leapp.libraries.actor import removeupgradeartifacts


@pytest.mark.parametrize(('exists', 'should_remove'), [
    (True, True),
    (False, False),
])
def test_remove_upgrade_artifacts(monkeypatch, exists, should_remove):

    called = [False]

    def mocked_run(cmd, *args, **kwargs):
        assert cmd[0] == 'rm'
        assert cmd[1] == '-rf'
        assert cmd[2] == removeupgradeartifacts.UPGRADE_ARTIFACTS_DIR
        called[0] = True
        return {'exit_code': 0, 'stdout': '', 'stderr': ''}

    monkeypatch.setattr(os.path, 'exists', lambda _: exists)
    monkeypatch.setattr(removeupgradeartifacts, 'run', mocked_run)

    removeupgradeartifacts.process()

    assert called[0] == should_remove



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class NetworkManagerConfig(Model):
    """The model contains NetworkManager configuration."""
    topic = SystemInfoTopic
    dhcp = fields.String(default='')



def getLockdownFirewallConfigCommand(root):
    for command in root.iter('command'):
        if 'name' in command.attrib and \
           '/usr/bin/firewall-config' in command.attrib['name']:
            return command.attrib['name']

    return ''


def getEbtablesTablesInUse(root):
    tables = []
    for rule in root.iter('rule'):
        if 'ipv' in rule.attrib and rule.attrib['ipv'] == 'eb' and \
           'table' in rule.attrib and rule.attrib['table'] not in tables:
            tables.append(rule.attrib['table'])

    for passthrough in root.iter('passthrough'):
        if 'ipv' in passthrough.attrib and passthrough.attrib['ipv'] == 'eb':
            rule = passthrough.text.split()
            try:
                i = rule.index('-t')
                if rule[i + 1] not in tables:
                    tables.append(rule[i + 1])
            except ValueError:
                pass

    return tables


def getIpsetTypesInUse(root):
    types = []
    for ipset in root.iter('ipset'):
        if 'type' in ipset.attrib and ipset.attrib['type'] not in types:
            types.append(ipset.attrib['type'])
    return types



from leapp.actors import Actor
from leapp.libraries.actor import checksystemdbrokensymlinks
from leapp.models import SystemdBrokenSymlinksSource, SystemdServicesInfoSource
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckSystemdBrokenSymlinks(Actor):
    """
    Check whether some systemd symlinks are broken

    If some systemd symlinks are broken, report them but do not inhibit the
    upgrade. The symlinks broken already before the upgrade will not be
    handled by the upgrade process anyhow. Two different reports are created:
    - symlinks which have the same filename as an existing enabled systemd
      service (the symlink doesn't point to an existing unit file, but the
      service is enabled)
    - broken symlinks which names do not correspond with any existing systemd
      unit file (typically when the service is removed but not disabled
      correctly)
    """

    name = 'check_systemd_broken_symlinks'
    consumes = (SystemdBrokenSymlinksSource, SystemdServicesInfoSource)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checksystemdbrokensymlinks.process()



from leapp.actors import Actor
from leapp.libraries.actor import checkcustommodifications
from leapp.models import CustomModifications, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckCustomModificationsActor(Actor):
    """
    Checks CustomModifications messages and produces a report about files in leapp directories that have been
    modified or newly added.
    """

    name = 'check_custom_modifications_actor'
    consumes = (CustomModifications,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        checkcustommodifications.report_any_modifications()



import pytest

from leapp.libraries.actor.cupsscanner import get_directive_value

testdata = (
    ('Include', 'Include smth', 'smth'),
    ('Include', 'something_else', None),
    ('Include', 'Include', ''),
    ('Include', '#Include smth', None),
    ('Include', '   Include smth', 'smth'),
    ('Include', '   Include smth anything', 'smth'),
)


@pytest.mark.parametrize('string, line, expected', testdata)
def test_get_directive_value(string, line, expected):

    value = get_directive_value(string, line)

    assert value == expected



from leapp.actors import Actor
from leapp.libraries.actor import scancryptopolicies
from leapp.models import CryptoPolicyInfo
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanCryptoPolicies(Actor):
    """
    Scan information about system wide set crypto policies including:
     * current crypto policy
     * installed custom crypto policies

    This information is, later in the process useful for the following:
     * copy the custom crypto policies files
     * notify user about the current setting and to review whether the policy still makes sense
       * it might be outdated and no longer meet the best security practices
       * if it is based on system policy such as DEFAULT, it might cause unexpected changes
    """

    name = 'scancryptopolicies'
    consumes = ()
    produces = (CryptoPolicyInfo,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scancryptopolicies.process()



import os

from leapp.exceptions import StopActorExecution
from leapp.libraries.stdlib import api
from leapp.models import BootContent


def remove_boot_files():
    boot_content_msgs = api.consume(BootContent)
    boot_content = next(boot_content_msgs, None)
    if list(boot_content_msgs):
        api.current_logger().warning('Unexpectedly received more than one BootContent message.')
    if not boot_content:
        api.current_logger().warning('Did not receive a message about the leapp-provided kernel and initramfs ->'
                                     ' Skipping removal of these files.')
        raise StopActorExecution
    for filepath in boot_content.kernel_path, boot_content.initram_path, boot_content.kernel_hmac_path:
        remove_file(filepath)


def remove_file(filepath):
    try:
        os.remove(filepath)
    except OSError as err:
        api.current_logger().error('Could not remove {0}: {1}.'.format(filepath, err))



from leapp.actors import Actor
from leapp.libraries.actor import multipathconfcheck
from leapp.models import MultipathConfFacts8to9
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class MultipathConfCheck8to9(Actor):
    """
    Checks if changes to the multipath configuration files are necessary
    for upgrading to RHEL9, and reports the results.
    """

    name = 'multipath_conf_check_8to9'
    consumes = (MultipathConfFacts8to9,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        facts = next(self.consume(MultipathConfFacts8to9), None)
        if facts is None:
            self.log.debug('Skipping execution. No MultipathConfFacts8to9 has '
                           'been produced')
            return
        multipathconfcheck.check_configs(facts)



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.stdlib import CalledProcessError, run
from leapp.models import Authselect, AuthselectDecision
from leapp.reporting import create_report, Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag

resources = [
    reporting.RelatedResource('package', 'authselect'),
    reporting.RelatedResource('package', 'authconfig'),
    reporting.RelatedResource('file', '/etc/nsswitch.conf')
]


class AuthselectApply(Actor):
    """
    Apply changes suggested by AuthselectScanner.

    If confirmed by admin in AuthselectDecision, call suggested authselect
    command to configure the system using this tool.
    """

    name = 'authselect_apply'
    consumes = (Authselect, AuthselectDecision,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        model = next(self.consume(Authselect))
        decision = next(self.consume(AuthselectDecision))

        if not decision.confirmed or model.profile is None:
            return

        command = ['authselect', 'select', '--force', model.profile] + model.features

        try:
            run(command)
        except CalledProcessError as err:
            create_report([  # pylint: disable-msg=too-many-arguments
                reporting.Title('Authselect call failed'),
                reporting.Summary(str(err)),
                reporting.Severity(reporting.Severity.MEDIUM),
                reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.TOOLS
                ]),
                reporting.Groups([
                    reporting.Groups.FAILURE
                ])
            ] + resources)  # pylint: disable-msg=too-many-arguments
            return

        try:
            run(['systemctl', 'enable', 'oddjobd.service'])
        except (OSError, CalledProcessError) as e:
            self.log.warning('Error enabling oddjobd.service: {}'.format(e))

        create_report([  # pylint: disable-msg=too-many-arguments
            reporting.Title('System was converted to authselect.'),
            reporting.Summary(
                'System was converted to authselect with the '
                'following call: "{}"'.format(' '.join(command))
            ),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.TOOLS
                ])
        ] + resources)  # pylint: disable-msg=too-many-arguments



import os
import tempfile

import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor.scancryptopolicies import (
    _get_name_from_file,
    find_rpm_untracked,
    read_current_policy,
    read_policy_dirs
)
from leapp.models import CustomCryptoPolicy, CustomCryptoPolicyModule

NOFILE = "/tmp/non-existing-file-should-not-really-be-here"


def test_get_name_from_file():
    assert _get_name_from_file("/path/name.extension") == "name"
    assert _get_name_from_file("/othername.e") == "othername"
    assert _get_name_from_file("/other.name.e") == "other.name"
    assert _get_name_from_file("/some/long/path/other.name") == "other"
    assert _get_name_from_file("/some/long/path/no_extension") == "no_extension"


def test_find_rpm_untracked(current_actor_context):
    # this is tracked
    files = ["/tmp/"]
    assert find_rpm_untracked(files) == []
    files = ["/etc/crypto-policies/config"]
    assert find_rpm_untracked(files) == []

    # the tempfile is not tracked by RPM
    with tempfile.NamedTemporaryFile(delete=False) as f:
        files = [f.name]
        assert find_rpm_untracked(files) == [f.name]

        # not existing files are ignored
        files = [NOFILE]
        assert find_rpm_untracked(files) == []

        # combinations should yield expected results too
        files = ["/tmp", f.name, NOFILE]
        assert find_rpm_untracked(files) == [f.name]
        # regardless the order
        files = [NOFILE, f.name, "/tmp"]
        assert find_rpm_untracked(files) == [f.name]


def test_read_current_policy():
    with pytest.raises(StopActorExecutionError):
        assert read_current_policy(NOFILE)

    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(b'DEFAULT:SHA1')
        f.flush()
        assert read_current_policy(f.name) == "DEFAULT:SHA1"

        f.seek(0)
        f.write(b'  DEFAULT:SHA1   \n\n   ')
        f.flush()
        assert read_current_policy(f.name) == "DEFAULT:SHA1"


def test_read_policy_dirs(current_actor_context):
    with tempfile.TemporaryDirectory() as dir1:
        # empty
        files = read_policy_dirs([dir1], CustomCryptoPolicy, ".pol")
        assert files == []

        # first policy module
        path1 = os.path.join(dir1, "policy.mpol")
        with open(path1, "x") as f:
            f.write('test')
        files = read_policy_dirs([dir1], CustomCryptoPolicy, ".pol")
        assert files == []
        files = read_policy_dirs([dir1], CustomCryptoPolicyModule, ".mpol")
        assert files == [CustomCryptoPolicyModule(name="policy", path=path1)]

        with tempfile.TemporaryDirectory() as dir2:
            files = read_policy_dirs([dir1], CustomCryptoPolicy, ".pol")
            assert files == []
            files = read_policy_dirs([dir1, dir2], CustomCryptoPolicyModule, ".mpol")
            assert files == [CustomCryptoPolicyModule(name="policy", path=path1)]

            # first policy file
            path2 = os.path.join(dir2, "mypolicy.pol")
            with open(path2, "x") as f:
                f.write('test2')
            # second policy file
            path3 = os.path.join(dir2, "other.pol")
            with open(path3, "x") as f:
                f.write('test3')

            files = read_policy_dirs([dir1, dir2], dict, ".pol")
            assert len(files) == 2
            assert dict(name="mypolicy", path=path2) in files
            assert dict(name="other", path=path3) in files
            files = read_policy_dirs([dir1, dir2], CustomCryptoPolicyModule, ".mpol")
            assert files == [CustomCryptoPolicyModule(name="policy", path=path1)]

        files = read_policy_dirs([dir1], CustomCryptoPolicy, ".pol")
        assert files == []
        files = read_policy_dirs([dir1], CustomCryptoPolicyModule, ".mpol")
        assert files == [CustomCryptoPolicyModule(name="policy", path=path1)]



from leapp import reporting
from leapp.libraries.common import rhsm
from leapp.libraries.stdlib import api
from leapp.models import RHSMInfo
from leapp.reporting import create_report

SCA_TEXT = "Content Access Mode is set to Simple Content Access"


def process():
    if not rhsm.skip_rhsm():
        for info in api.consume(RHSMInfo):
            if not info.attached_skus and not info.sca_detected:
                create_report([
                    reporting.Title('The system is not registered or subscribed.'),
                    reporting.Summary(
                        'The system has to be registered and subscribed to be able to proceed'
                        ' with the upgrade, unless the --no-rhsm option is specified when'
                        ' executing leapp.'
                    ),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.SANITY]),
                    reporting.Groups([reporting.Groups.INHIBITOR]),
                    reporting.Remediation(
                        hint='Register your system with the subscription-manager tool and attach'
                             ' proper SKUs to be able to proceed the upgrade or use the --no-rhsm'
                             ' leapp option if you want to provide target repositories by yourself.'),
                    reporting.RelatedResource('package', 'subscription-manager')
                ])



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckDosfstools(Actor):
    """
    Check if dosfstools is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkdosfstools'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'dosfstools'):
            create_report([
                reporting.Title('Dosfstools incompatible changes in the next major version'),
                reporting.Summary(
                    'The automatic alignment of data clusters that was added in 3.0.8 and broken for '
                    'FAT32 starting with 3.0.20 has been reinstated. If you need to create file systems '
                    'for finicky devices that have broken FAT implementations use the option -a to '
                    'disable alignment.\n'
                    'The fsck.fat now defaults to interactive repair mode which previously had to be '
                    'selected with the -r option.\n'
                ),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Groups([
                        reporting.Groups.FILESYSTEM,
                        reporting.Groups.TOOLS
                ]),
                reporting.Remediation(hint='Please update your scripts to be compatible with the changes.'),
                reporting.RelatedResource('package', 'dosfstools')
            ])



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import CupsChangedFeatures


def _get_input_model(model):
    """
    Gets data model from an actor.

    :param obj model: object of model which data will be consumed
    """
    return next(api.consume(model), None)


def check_interface_scripts(facts, report_func):
    """
    Checks if the data model tells interface scripts are used
    and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('CUPS no longer supports usage of interface scripts')
    summary = ('Interface scripts are no longer supported due to '
               'security issues - an attacker could provide '
               'malicious script which will be run during printing.')
    hint = ('Install the queue with PPD file for the printer '
            'if available or install the queue with generic PPD, '
            'add *cupsFilter2 directive into PPD of installed '
            'queue (in /etc/cups/ppd) and reinstall the queue with modified PPD. '
            'The interface script needs to have permissions 750 and '
            'ownership root:lp. How to write *cupsFilter2 keyword '
            'is described at https://www.cups.org/doc/spec-ppd.html#cupsFilter2 '
            'and the script needs to be put into /usr/lib/cups/filter '
            'or you need to use an absolute path to the script '
            'in *cupsFilter2 directive.')
    if facts.interface:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([reporting.Groups.DRIVERS]),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Remediation(hint=hint),
            reporting.ExternalLink(
                title='Upstream documentation for the cupsFilter2 PPD keyword',
                url='https://www.cups.org/doc/spec-ppd.html#cupsFilter2'
            )
        ]

        report_func(args)


def check_include_directive(facts, report_func):
    """
    Checks if the data model tells include directives are used
    and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('CUPS no longer supports usage of Include directive')
    summary = ('Include directive was removed due to security reasons. '
               'Contents of found included files will be appended to '
               'cupsd.conf')
    if facts.include:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([reporting.Groups.SERVICES]),
            reporting.Severity(reporting.Severity.MEDIUM),
        ] + [reporting.RelatedResource('file', f) for f in facts.include_files]

        report_func(args)


def check_printcap_directive(facts, report_func):
    """
    Checks if the data model tells printcapformat directive is used
    and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('PrintcapFormat directive is no longer in cupsd.conf')
    summary = (
        'The directive was moved into /etc/cups/cups-files.conf '
        'because it is deprecated. This will be handled automatically during '
        'the upgrade process.'
    )
    if facts.printcap:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([reporting.Groups.SERVICES]),
            reporting.Severity(reporting.Severity.LOW),
            reporting.RelatedResource('file', '/etc/cups/cupsd.conf'),
            reporting.RelatedResource('file', '/etc/cups/cups-files.conf')
        ]

        report_func(args)


def check_env_directives(facts, report_func):
    """
    Checks if the data model tells PassEnv/SetEnv directives are used
    and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('PassEnv/SetEnv directives are no longer in cupsd.conf')
    summary = (
        'The directives were moved into /etc/cups/cups-files.conf '
        'due to security reasons. '
        'This will be handled automatically during the upgrade process.'
    )
    if facts.env:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([reporting.Groups.SERVICES]),
            reporting.Severity(reporting.Severity.LOW),
            reporting.RelatedResource('file', '/etc/cups/cupsd.conf'),
            reporting.RelatedResource('file', '/etc/cups/cups-files.conf')
        ]

        report_func(args)


def check_certkey_directives(facts, report_func):
    """
    Checks if the data model tells ServerKey/ServerCertificate directives
    are used and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('ServerKey/ServerCertificate directives are substituted '
             'by ServerKeychain directive')
    summary = (
        'The directives were substituted by ServerKeychain directive, '
        'which now takes a directory as value (/etc/cups/ssl is default). '
        'The previous directives took a file as value. '
        'The migration script will copy the files specified in '
        'directive values into /etc/cups/ssl directory '
        'if they are not there already. '
        'This will be handled automatically during the upgrade process.'
    )
    if facts.certkey:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([reporting.Groups.SERVICES,
                              reporting.Groups.AUTHENTICATION]),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.RelatedResource('file', '/etc/cups/cups-files.conf')
        ]

        report_func(args)


def check_digest_values(facts, report_func):
    """
    Checks if the data model tells Digest/BasicDigest values
    of AuthType/DefaultAuthType directives are used
    and produces a report.

    :param obj facts: model object containing info about CUPS configuration
    :param func report_func: creates report
    """
    title = ('CUPS no longer supports Digest and BasicDigest '
             'directive values')
    summary = (
        'Digest and BasicDigest directive values were removed '
        'due to deprecation. '
        'The Basic authentication with TLS encryption will be '
        'set automatically during the upgrade process. '
        'The version of the used TLS is by default dependent on the set system '
        'crypto policies.'
    )
    # NOTE: the remediation instructions are missing as we do not have any
    # doc covering that, mainly because of this is expected to be very rare
    # at all. People usually do not use Digest & BasicDigest.
    if facts.digest:
        args = [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Groups([
                reporting.Groups.AUTHENTICATION,
                reporting.Groups.SECURITY,
                reporting.Groups.SERVICES,
            ]),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.RelatedResource('file', '/etc/cups/cupsd.conf')
        ]

        report_func(args)


def make_reports(consume_function=_get_input_model,
                 report_func=reporting.create_report,
                 debug_log=api.current_logger().debug):
    """
    Creates reports if needed

    :param func consume_function: gets data model from an actor
    :param func report_func: creates report
    """
    facts = consume_function(CupsChangedFeatures)

    if facts:
        check_interface_scripts(facts, report_func)
        check_include_directive(facts, report_func)
        check_printcap_directive(facts, report_func)
        check_env_directives(facts, report_func)
        check_certkey_directives(facts, report_func)
        check_digest_values(facts, report_func)
    else:
        debug_log('No facts gathered about CUPS - skipping reports.')



import pytest

from leapp.libraries.actor import checksaphana
from leapp.libraries.common import testutils
from leapp.libraries.common.config import architecture, version
from leapp.models import SapHanaManifestEntry

SAPHANA1_MANIFEST = '''comptype: HDB
keyname: HDB
keycaption: SAP HANA DATABASE
supported-phases: prepare,offline,online
requires-restart: system
keyvendor: sap.com
release: 1.00
rev-number: 122
rev-patchlevel: 33
sp-number: 122
sp-patchlevel: 33
makeid: 7407089
date: 2020-10-08 16:23:37
platform: linuxx86_64
hdb-state: RAMP
fullversion: 1.00.122.33 Build 1602166441-1530
'''


SAPHANA2_MANIFEST = '''comptype: HDB
keyname: HDB
keycaption: SAP HANA DATABASE
supported-phases: prepare,offline,configure,online
requires-restart: system
keyvendor: sap.com
release: 2.00
rev-number: 053
rev-patchlevel: 00
sp-number: 053
sp-patchlevel: 00
max_rel2.0_sps2_rev-number: 24
max_rel2.0_sps2_patchlevel: 10
max_rel2.0_sps3_rev-number: 37
max_rel2.0_sps3_patchlevel: 07
max_rel2.0_sps4_rev-number: 48
max_rel2.0_sps4_patchlevel: 02
date: 2020-11-11 12:12:22
platform: linuxx86_64
hdb-state: RAMP
fullversion: 2.00.053.00 Build 1605092543-1530
'''


SAPHANA2_LOW_VERSION_MANIFEST = '''comptype: HDB
keyname: HDB
keycaption: SAP HANA DATABASE
supported-phases: prepare,offline,configure,online
requires-restart: system
keyvendor: sap.com
release: 2.00
rev-number: 040
rev-patchlevel: 00
sp-number: 040
sp-patchlevel: 00
max_rel2.0_sps2_rev-number: 24
max_rel2.0_sps2_patchlevel: 10
max_rel2.0_sps3_rev-number: 37
max_rel2.0_sps3_patchlevel: 07
date: 2020-11-11 12:12:22
platform: linuxx86_64
hdb-state: RAMP
fullversion: 2.00.040.00 Build 1605092543-1530
'''


def _report_has_pattern(report, pattern):
    return pattern in report[0].to_dict().get('title', '')


EXPECTED_TITLE_PATTERNS = {
    'running': lambda report: _report_has_pattern(report, 'running SAP HANA'),
    'v1': lambda report: _report_has_pattern(report, 'Found SAP HANA 1'),
    'low': lambda report: _report_has_pattern(report, 'SAP HANA needs to be updated before the RHEL upgrade'),
}


def list_clear(l):
    del l[:]


def _parse_manifest_data(manifest):
    result = []
    for line in manifest.split('\n'):
        line = line.strip()
        if not line:
            continue
        # If there is a ValueError during the split, just burn with fire in this test
        key, value = line.split(': ', 1)
        result.append(SapHanaManifestEntry(key=key, value=value.strip()))
    return result


class MockSapHanaInstanceInfo(object):
    def __init__(self, name, number, path, admin, manifest_data, running=True):
        self.manifest = _parse_manifest_data(manifest_data)
        self.name = name
        self.instance_number = number
        self.path = path
        self.running = running
        self.admin = admin


def _gen_instance_info(name, manifest_data, index, running=True):
    return MockSapHanaInstanceInfo(
        name=name,
        number='{:02}'.format(index),
        path='/hana/shared/{name}/HDB{number:02}'.format(name=name, number=index),
        admin='{name}adm'.format(name=name.lower()),
        manifest_data=manifest_data,
        running=running
    )


class MockSapHanaInfo(object):
    def __init__(self, v1names, v2names, v2lownames, running=None):
        self.installed = bool(v1names or v2names or v2lownames)
        self.running = running if running is not None else self.installed
        self.instances = [_gen_instance_info(name,
                                             SAPHANA1_MANIFEST,
                                             index,
                                             running=running)
                          for index, name in enumerate(v1names)]
        self.instances += [_gen_instance_info(name,
                                              SAPHANA2_MANIFEST,
                                              index + len(v1names),
                                              running=running)
                           for index, name in enumerate(v2names)]
        self.instances += [_gen_instance_info(name,
                                              SAPHANA2_LOW_VERSION_MANIFEST,
                                              index + len(v1names) + len(v2names),
                                              running=running)
                           for index, name in enumerate(v2lownames)]


def _report_collector(reports):
    def _mock_report(args):
        reports.append(args)
    return _mock_report


def _consume_mock_sap_hana_info(v1names=(), v2names=(), v2lownames=(), running=True):
    def _consume(*models):
        return iter([MockSapHanaInfo(v1names, v2names, v2lownames, running=running)])
    return _consume


class MockSAPHanaVersionInstance(object):
    def __init__(self, major, rev, patchlevel):
        self.name = "TestName"

        def _kv(k, v):
            return SapHanaManifestEntry(key=k, value=v)
        self.manifest = [
            _kv('release', '{major}.00'.format(major=major)),
            _kv('rev-number', '{rev:03}'.format(rev=rev)),
            _kv('rev-patchlevel', '{pl:02}'.format(pl=patchlevel)),
            _kv('sp-number', '{rev:03}'.format(rev=rev)),
            _kv('sp-patchlevel', '{pl:02}'.format(pl=patchlevel)),
        ]


@pytest.mark.parametrize(
    'major,rev,patchlevel,result', (
        (2, 52, 0, True),
        (2, 52, 1, True),
        (2, 52, 2, True),
        (2, 53, 0, True),
        (2, 60, 0, True),
        (2, 48, 2, True),
        (2, 48, 1, False),
        (2, 48, 0, False),
        (2, 38, 2, False),
        (2, 49, 0, True),
    )
)
def test_checksaphana__fullfills_rhel86_hana_min_version(monkeypatch, major, rev, patchlevel, result):
    monkeypatch.setattr(version, 'get_target_major_version', lambda: '8')
    monkeypatch.setattr(version, 'get_target_version', lambda: '8.6')
    monkeypatch.setattr(checksaphana, 'SAP_HANA_RHEL86_REQUIRED_PATCH_LEVELS', ((4, 48, 2), (5, 52, 0)))
    assert checksaphana._fullfills_hana_min_version(
        MockSAPHanaVersionInstance(
            major=major,
            rev=rev,
            patchlevel=patchlevel,
        )
    ) == result


@pytest.mark.parametrize(
    'major,rev,patchlevel,result', (
        (2, 59, 4, True),
        (2, 59, 5, True),
        (2, 59, 6, True),
        (2, 60, 0, False),
        (2, 61, 0, False),
        (2, 62, 0, False),
        (2, 63, 2, True),
        (2, 48, 1, False),
        (2, 48, 0, False),
        (2, 59, 0, False),
        (2, 59, 1, False),
        (2, 59, 2, False),
        (2, 59, 3, False),
        (2, 38, 2, False),
        (2, 64, 0, True),
    )
)
def test_checksaphana__fullfills_hana_rhel90_min_version(monkeypatch, major, rev, patchlevel, result):
    monkeypatch.setattr(version, 'get_target_major_version', lambda: '9')
    monkeypatch.setattr(version, 'get_target_version', lambda: '9.0')
    monkeypatch.setattr(checksaphana, 'SAP_HANA_RHEL90_REQUIRED_PATCH_LEVELS', ((5, 59, 4), (6, 63, 0)))
    assert checksaphana._fullfills_hana_min_version(
        MockSAPHanaVersionInstance(
            major=major,
            rev=rev,
            patchlevel=patchlevel,
        )
    ) == result


@pytest.mark.parametrize('flavour', ('default', 'saphana'))
@pytest.mark.parametrize('version,arch,inhibitor_expected', (
    ('8.6', architecture.ARCH_X86_64, False),
    ('8.6', architecture.ARCH_PPC64LE, True),
    ('8.6', architecture.ARCH_ARM64, True),
    ('8.6', architecture.ARCH_S390X, True),

    ('9.0', architecture.ARCH_X86_64, False),
    ('9.0', architecture.ARCH_PPC64LE, False),
    ('9.0', architecture.ARCH_ARM64, True),
    ('9.0', architecture.ARCH_S390X, True),
))
def test_checksaphsana_test_arch(monkeypatch, flavour, version, arch, inhibitor_expected):
    reports = []
    monkeypatch.setattr(checksaphana.reporting, 'create_report', _report_collector(reports))
    curr_actor_mocked = testutils.CurrentActorMocked(arch=arch, flavour=flavour, dst_ver=version)
    monkeypatch.setattr(checksaphana.api, 'current_actor', curr_actor_mocked)
    checksaphana.perform_check()
    if flavour == 'saphana' and inhibitor_expected:
        # the system has SAP HANA but unsupported target arch
        assert reports and len(reports) == 1
        assert 'x86_64' in reports[0][1].to_dict()['summary']
        if version[0] == '9':
            assert 'ppc64le' in reports[0][1].to_dict()['summary']
    elif flavour != 'saphana' or not inhibitor_expected:
        assert not reports


def test_checksaphana_perform_check(monkeypatch):
    v1names = ('ABC', 'DEF', 'GHI')
    v2names = ('JKL', 'MNO', 'PQR', 'STU')
    v2lownames = ('VWX', 'YZA')
    reports = []
    monkeypatch.setattr(checksaphana, 'SAP_HANA_RHEL86_REQUIRED_PATCH_LEVELS', ((4, 48, 2), (5, 52, 0)))
    monkeypatch.setattr(version, 'get_target_major_version', lambda: '8')
    monkeypatch.setattr(version, 'get_target_version', lambda: '8.6')
    monkeypatch.setattr(checksaphana.reporting, 'create_report', _report_collector(reports))
    monkeypatch.setattr(checksaphana.api, 'consume', _consume_mock_sap_hana_info(
        v1names=v1names, v2names=v2names, v2lownames=v2lownames, running=True))

    list_clear(reports)
    monkeypatch.setattr(checksaphana.api,
                        'current_actor',
                        testutils.CurrentActorMocked(arch=architecture.ARCH_X86_64))
    checksaphana.perform_check()
    assert not reports

    monkeypatch.setattr(checksaphana.api,
                        'current_actor',
                        testutils.CurrentActorMocked(arch=architecture.ARCH_X86_64, flavour='saphana'))
    checksaphana.perform_check()
    assert reports
    # Expected 3 reports due to v1names + v2lownames + running
    assert len(reports) == 3
    # Verifies that all expected title patterns are within the reports and not just coincidentally 3
    assert all([any([pattern(report) for report in reports]) for pattern in EXPECTED_TITLE_PATTERNS.values()])

    list_clear(reports)
    monkeypatch.setattr(checksaphana.api, 'consume', _consume_mock_sap_hana_info(
        v1names=v1names, v2names=v2names, v2lownames=v2lownames, running=False))
    checksaphana.perform_check()
    assert reports
    # Expected 2 reports due to v1names + v2lownames
    assert len(reports) == 2
    # Verifies that all expected title patterns are within the reports and not just coincidentally 2
    assert all([any([EXPECTED_TITLE_PATTERNS[pattern](report) for report in reports]) for pattern in ['v1', 'low']])



import os
import re

SendmailConfDir = '/etc/mail'
SendmailConfFiles = ['sendmail.cf', 'sendmail.mc', 'submit.cf', 'submit.mc']
# false positives blacklist
rfp = re.compile(r'(^\s*RIPv6:::1\b)|(@\s+\[IPv6:::1\]\s+>)')


def get_conf_files():
    conf_files = [os.path.join(SendmailConfDir, f) for f in SendmailConfFiles]
    return conf_files


def check_false_positives(filename, line):
    return filename in ['sendmail.cf', 'submit.cf'] and rfp.search(line) is not None


def check_files_for_compressed_ipv6():
    conf_files = get_conf_files()
    migrate_files = []
    files = [os.path.join(SendmailConfDir, re.sub(r'\.db$', '', f)) for f in os.listdir(SendmailConfDir)
             if f.endswith('.db')] + conf_files
    regex = re.compile(r'IPv6:[0-9a-fA-F:]*::')
    for filename in files:
        if not os.path.exists(filename):
            continue
        with open(filename) as file_check:
            for line in file_check:
                if regex.search(line) and not check_false_positives(os.path.basename(filename), line):
                    migrate_files.append(filename)
                    break
    return migrate_files



from leapp.actors import Actor
from leapp.libraries.actor import scancpu
from leapp.models import CPUInfo, DetectedDeviceOrDriver, DeviceDriverDeprecationData
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanCPU(Actor):
    """Scan CPUs of the machine."""

    name = 'scancpu'
    consumes = (DeviceDriverDeprecationData,)
    produces = (CPUInfo, DetectedDeviceOrDriver)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self, *args, **kwargs):
        scancpu.process()



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class User(Model):
    topic = SystemFactsTopic

    name = fields.String()
    uid = fields.Integer()
    gid = fields.Integer()
    home = fields.String()


class UsersFacts(Model):
    topic = SystemFactsTopic

    users = fields.List(fields.Model(User))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class BindConfigIssuesModel(Model):
    """
    Problematic files with statements, which are problematic
    """

    topic = SystemInfoTopic
    path = fields.String()  # path to problematic file
    statements = fields.List(fields.String())  # list of offending statements


class BindFacts(Model):
    """
    Whole facts related to BIND configuration
    """

    topic = SystemInfoTopic

    # Detected configuration files via includes
    config_files = fields.List(fields.String())

    # Files modified by update
    modified_files = fields.List(fields.String())

    # Only issues detected.
    # unsupported dnssec-lookaside statements with old values
    # found in list of files. List of files, where unsupported
    # statements were found. Context not yet provided
    dnssec_lookaside = fields.Nullable(fields.List(fields.Model(BindConfigIssuesModel)))

    # Missing listen-on-v6 option
    listen_on_v6_missing = fields.Boolean(default=False)



import pytest

from leapp.libraries.actor.cupsfiltersmigrate import NEW_MACROS, update_config


def _gen_append_str(list_out=None):
    """
    Just helper function to generate string expected to be added for an input (see testdata) for testing.

    :param list list_out: None, [0], [1], [0,1] - no more expected vals,
                          which represents what macros should be appended
                          in output
    """
    if not list_out:
        return ''
    _out_list = ('LocalQueueNamingRemoteCUPS RemoteName', 'CreateIPPPrinterQueues All')
    output = ['# content added by Leapp']
    for i in list_out:
        output.append(_out_list[i])
    # ensure the extra NL is before the string and the empty NL is in the end
    # of the string (/file) as well
    return '\n{}\n'.format('\n'.join(output))


testdata = (
    ('\n',
        _gen_append_str([0, 1])),
    ('bleblaba\n',
        _gen_append_str([0, 1])),
    ('fdnfdf\n# LocalQueueNamingRemoteCUPS RemoteName\n',
        _gen_append_str([0, 1])),
    ('fdnfdf\nfoo # LocalQueueNamingRemoteCUPS RemoteName\n',
        _gen_append_str([0, 1])),
    ('fdnfdf\n# LocalQueueNamingRemoteCUPS Bar\n',
        _gen_append_str([0, 1])),
    ('fdnfdf\n  # LocalQueueNamingRemoteCUPS Bar\n',
        _gen_append_str([0, 1])),
    ('fdnfdf\nLocalQueueNamingRemoteCUPS RemoteName\n',
        _gen_append_str([1])),
    ('fdnfdf\n  LocalQueueNamingRemoteCUPS RemoteName\n',
        _gen_append_str([1])),
    ('fdnfdf\nLocalQueueNamingRemoteCUPS Bar\n',
        _gen_append_str([1])),
    ('fnfngbfg\nCreateIPPPrinterQueues All\n',
        _gen_append_str([0])),
    ('fnfngbfg\nCreateIPPPrinterQueues Foo\n',
        _gen_append_str([0])),
    ('fnfngbfg\n  CreateIPPPrinterQueues Foo\n',
        _gen_append_str([0])),
    ('CreateIPPPrinterQueues All\nLocalQueueNamingRemoteCUPS RemoteName\n',
        _gen_append_str()),
    ('CreateIPPPrinterQueues Foo\nLocalQueueNamingRemoteCUPS Bar\n',
        _gen_append_str()),
    ('foo\nCreateIPPPrinterQueues Foo\nLocalQueueNamingRemoteCUPS Bar\nFoobar\n',
        _gen_append_str()),
    ('foo\nCreateIPPPrinterQueues Foo\n# LocalQueueNamingRemoteCUPS Bar\nFoobar\n',
        _gen_append_str([0]))
)


class MockFile(object):
    def __init__(self, path, content=None):
        self.path = path
        self.content = content
        self.error = False

    def append(self, path, content):
        if path != self.path:
            self.error = True
        if not self.error:
            self.content += content
            return self.content
        raise IOError('Error during writing to file: {}.'.format(path))

    def exists(self, path, macro):
        for line in self.content.split('\n'):
            if line.lstrip().startswith(macro) and self.path == path:
                return True
        return False


def test_update_config_file_errors():
    path = 'foo'

    f = MockFile(path, content='')

    with pytest.raises(IOError):
        update_config('bar', f.exists, f.append)

    assert f.content == ''


@pytest.mark.parametrize('content,expected', testdata)
def test_update_config_append_into_file(content, expected):
    path = 'bar'
    f = MockFile(path, content)

    update_config(path, f.exists, f.append)

    assert f.content == content + expected



import itertools
import os
from collections import namedtuple

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import rhsm, rhui
from leapp.libraries.common.config import version
from leapp.libraries.stdlib import api
from leapp.models import (
    CopyFile,
    DNFPluginTask,
    InstalledRPM,
    RHUIInfo,
    RpmTransactionTasks,
    TargetRHUIPostInstallTasks,
    TargetRHUIPreInstallTasks,
    TargetRHUISetupInfo,
    TargetUserSpacePreupgradeTasks
)

MatchingSetup = namedtuple('MatchingSetup', ['family', 'description'])


def into_set(pkgs):
    if isinstance(pkgs, set):
        return pkgs
    if isinstance(pkgs, str):
        return {pkgs}
    return set(pkgs)


def find_rhui_setup_matching_src_system(installed_pkgs, rhui_map):
    src_ver = version.get_source_major_version()
    arch = api.current_actor().configuration.architecture

    matching_setups = []
    for rhui_family, family_setups in rhui_map.items():
        if rhui_family.arch != arch:
            continue

        for setup in family_setups:
            if setup.os_version != src_ver:
                continue
            if setup.clients.issubset(installed_pkgs):
                matching_setups.append(MatchingSetup(family=rhui_family, description=setup))

    if not matching_setups:
        return None

    # In case that a RHUI variant uses a combination of clients identify the maximal client set
    matching_setups_by_size = sorted(matching_setups, key=lambda match: -len(match.description.clients))

    match = matching_setups_by_size[0]  # Matching setup with the highest number of clients
    if len(matching_setups) == 1:
        return match

    if len(matching_setups_by_size[0].description.clients) == len(matching_setups_by_size[1].description.clients):
        # Should not happen as no cloud providers use multi-client setups (at the moment)
        msg = 'Could not identify the source RHUI setup (ambiguous setup)'

        variant_detail_table = {
           rhui.RHUIVariant.ORDINARY: '',
           rhui.RHUIVariant.SAP: ' for SAP',
           rhui.RHUIVariant.SAP_APPS: ' for SAP Applications',
           rhui.RHUIVariant.SAP_HA: ' for SAP HA',
        }

        match0 = matching_setups_by_size[0]
        variant0_detail = variant_detail_table[match0.family.variant]
        clients0 = ' '.join(match0.description.clients)

        match1 = matching_setups_by_size[1]
        variant1_detail = variant_detail_table[match1.family.variant]
        clients1 = ' '.join(match1.description.clients)

        details = ('Leapp uses client-based identification of the used RHUI setup in order to determine what the '
                   'target RHEL content should be. According to the installed RHUI clients the system should be '
                   'RHEL {os_major}{variant0_detail} ({provider0}) (identified by clients {clients0}) but also '
                   'RHEL {os_major}{variant1_detail} ({provider1}) (identified by clients {clients1}).')
        details = details.format(os_major=version.get_source_major_version(),
                                 variant0_detail=variant0_detail, clients0=clients0, provider0=match0.family.provider,
                                 variant1_detail=variant1_detail, clients1=clients1, provider1=match1.family.provider)

        raise StopActorExecutionError(message=msg, details={'details': details})

    return match


def determine_target_setup_desc(cloud_map, rhui_family):
    variant_setups = cloud_map[rhui_family]
    target_major = version.get_target_major_version()

    for setup in variant_setups:
        if setup.os_version == target_major:
            return setup
    return None


def inhibit_if_leapp_pkg_to_access_target_missing(installed_pkgs, rhui_family, target_setup_desc):
    pkg_name = target_setup_desc.leapp_pkg

    if pkg_name not in installed_pkgs:
        summary = 'On {provider} the "{pkg}" is required to perform an in-place upgrade'
        summary = summary.format(provider=rhui_family.provider, pkg=pkg_name)
        reporting.create_report([
            reporting.Title('Package "{}" is not installed'.format(pkg_name)),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.RelatedResource('package', pkg_name),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Groups([reporting.Groups.PUBLIC_CLOUD, reporting.Groups.RHUI]),
            reporting.Remediation(commands=[['yum', 'install', '-y', pkg_name]])
        ])
        return True
    return False


def stop_due_to_unknown_target_system_setup(rhui_family):
    msg = 'Failed to identify target RHUI setup'
    variant_detail = ' ({rhui_family.variant})' if rhui_family.variant != rhui.RHUIVariant.ORDINARY else ''
    details = ('Leapp successfully identified the current RHUI setup as a system provided by '
               '{provider}{variant_detail}, but it failed to determine'
               ' equivalent RHUI setup for the target OS.')
    details = details.format(provider=rhui_family.provider, variant_detail=variant_detail)
    raise StopActorExecutionError(message=msg, details={'details': details})


def customize_rhui_setup_for_gcp(rhui_family, setup_info):
    if not rhui_family.provider == rhui.RHUIProvider.GOOGLE:
        return

    # The google-cloud.repo repofile provides the repository containing the target clients. However, its repoid is the
    # same across all rhel versions, therefore, we need to remove the source google-cloud.repo to enable
    # correct target one.
    setup_info.preinstall_tasks.files_to_remove.append('/etc/yum.repos.d/google-cloud.repo')


def customize_rhui_setup_for_aws(rhui_family, setup_info):
    if rhui_family.provider != rhui.RHUIProvider.AWS:
        return

    target_version = version.get_target_major_version()
    if target_version == '8':
        return  # The rhel8 plugin is packed into leapp-rhui-aws as we need python2 compatible client

    amazon_plugin_copy_task = CopyFile(src='/usr/lib/python3.9/site-packages/dnf-plugins/amazon-id.py',
                                       dst='/usr/lib/python3.6/site-packages/dnf-plugins/')
    setup_info.postinstall_tasks.files_to_copy.append(amazon_plugin_copy_task)


def produce_rhui_info_to_setup_target(rhui_family, source_setup_desc, target_setup_desc):
    rhui_files_location = os.path.join(api.get_common_folder_path('rhui'), rhui_family.client_files_folder)

    files_to_access_target_client_repo = []
    for filename, target_path in target_setup_desc.mandatory_files:
        src_path = os.path.join(rhui_files_location, filename)
        files_to_access_target_client_repo.append(CopyFile(src=src_path, dst=target_path))

    for filename, target_path in target_setup_desc.optional_files:
        src_path = os.path.join(rhui_files_location, filename)

        if not os.path.exists(src_path):
            msg = "Optional file {} is present, will be used to setup target RHUI."
            api.current_logger().debug(msg.format(src_path))
            continue

        files_to_access_target_client_repo.append(CopyFile(src=src_path, dst=target_path))

    preinstall_tasks = TargetRHUIPreInstallTasks(files_to_copy_into_overlay=files_to_access_target_client_repo)

    files_supporting_client_operation = sorted(
        os.path.join(rhui_files_location, file) for file in target_setup_desc.files_supporting_client_operation
    )

    target_client_setup_info = TargetRHUISetupInfo(
        preinstall_tasks=preinstall_tasks,
        postinstall_tasks=TargetRHUIPostInstallTasks(),
        files_supporting_client_operation=files_supporting_client_operation
    )

    customize_rhui_setup_for_gcp(rhui_family, target_client_setup_info)
    customize_rhui_setup_for_aws(rhui_family, target_client_setup_info)

    rhui_info = RHUIInfo(
        provider=rhui_family.provider.lower(),
        variant=rhui_family.variant,
        src_client_pkg_names=sorted(source_setup_desc.clients),
        target_client_pkg_names=sorted(target_setup_desc.clients),
        target_client_setup_info=target_client_setup_info
    )
    api.produce(rhui_info)


def produce_rpms_to_install_into_target(source_setup, target_setup):
    to_install = sorted(target_setup.clients - source_setup.clients)
    to_remove = sorted(source_setup.clients - target_setup.clients)

    api.produce(TargetUserSpacePreupgradeTasks(install_rpms=sorted(target_setup.clients)))
    if to_install or to_remove:
        api.produce(RpmTransactionTasks(to_install=to_install, to_remove=to_remove))


def inform_about_upgrade_with_rhui_without_no_rhsm():
    if not rhsm.skip_rhsm():
        reporting.create_report([
            reporting.Title('Upgrade initiated with RHSM on public cloud with RHUI infrastructure'),
            reporting.Summary(
                'Leapp detected this system is on public cloud with RHUI infrastructure '
                'but the process was initiated without "--no-rhsm" command line option '
                'which implies RHSM usage (valid subscription is needed).'
            ),
            reporting.Severity(reporting.Severity.INFO),
            reporting.Groups([reporting.Groups.PUBLIC_CLOUD]),
        ])
        return True
    return False


def process():
    installed_rpm = itertools.chain(*[installed_rpm_msg.items for installed_rpm_msg in api.consume(InstalledRPM)])
    installed_pkgs = {rpm.name for rpm in installed_rpm}

    src_rhui_setup = find_rhui_setup_matching_src_system(installed_pkgs, rhui.RHUI_SETUPS)
    if not src_rhui_setup:
        return
    api.current_logger().debug("The RHUI family of the source system is {}".format(src_rhui_setup.family))

    target_setup_desc = determine_target_setup_desc(rhui.RHUI_SETUPS, src_rhui_setup.family)

    if not target_setup_desc:
        # We know that we are on RHUI because we have identified what RHUI variant it is, but we don't know how does
        # the target system look like. Likely, our knowledge of what RHUI setups are there (RHUI_SETUPS) is incomplete.
        stop_due_to_unknown_target_system_setup(src_rhui_setup.family)
        return

    if inform_about_upgrade_with_rhui_without_no_rhsm():
        return

    if inhibit_if_leapp_pkg_to_access_target_missing(installed_pkgs, src_rhui_setup.family, target_setup_desc):
        return

    # Instruction on how to access the target content
    produce_rhui_info_to_setup_target(src_rhui_setup.family, src_rhui_setup.description, target_setup_desc)

    produce_rpms_to_install_into_target(src_rhui_setup.description, target_setup_desc)

    if src_rhui_setup.family.provider == rhui.RHUIProvider.AWS:
        # We have to disable Amazon-id plugin in the initramdisk phase as there is no network
        api.produce(DNFPluginTask(name='amazon-id', disable_in=['upgrade']))



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class PCIDevice(Model):
    """
    Model to represent a PCI Device.

    There is the following match between model parameters and
    the fields of the output of a shell command `lspci -vmmk`

    slot - 'Slot',
    dev_cls - 'Class',
    vendor - 'Vendor',
    name - 'Device',
    subsystem_vendor - 'SVendor',
    subsystem_name - 'SDevice',
    physical_slot - 'PhySlot',
    rev - 'Rev',
    progif - 'ProgIf',
    driver - 'Driver',
    modules - 'Module',
    numa_node - 'NUMANode',

    pci_id - represents the numeric identification of the device, formed as
        string concatenating of the numeric device identifiers for fields
        Vendor, Device, SVendor, SDevice (output
        of a shell command `lspci -vmmkn`) with the `:` delimiter.
        For example:
        one device from output of `lspci -vmmkn` is:

        ```
        Slot:	04:00.0
        Class:	0880
        Vendor:	8086
        Device:	15bf
        SVendor:	17aa
        SDevice:	2279
        Rev:	01
        Driver:	thunderbolt
        Module:	thunderbolt
        ```

        then
        pci_id == "8086:15bf:17aa:2279"
    """
    topic = SystemInfoTopic

    slot = fields.String()
    dev_cls = fields.String()
    vendor = fields.String()
    name = fields.String()
    subsystem_vendor = fields.Nullable(fields.String())
    subsystem_name = fields.Nullable(fields.String())
    physical_slot = fields.Nullable(fields.String())
    rev = fields.Nullable(fields.String())
    progif = fields.Nullable(fields.String())
    driver = fields.Nullable(fields.String())
    modules = fields.Nullable(fields.List(fields.String()))
    numa_node = fields.Nullable(fields.String())
    pci_id = fields.String()


class PCIDevices(Model):
    topic = SystemInfoTopic

    devices = fields.List(fields.Model(PCIDevice))


class RestrictedPCIDevice(Model):
    """
    Model to represent known restrictions of the given PCI devices.


    pci_id - unsupported pci_ids. It has the following
        structure: {Vendor}:{Device}:{SVendor}:{SDevice}, where all these 4
        parameters are numeric ids (see shell command spci -vmmkn). If SVendor
        and SDevice fields do not exist, then pci_id has the following structure:
        {Vendor}:{Device}.
    driver_name - the name of restricted driver
    device_name - the name of restricted device
    supported_{rhel_version} - 1 is supported on the given RHEL, 0 - not
        supported
    available_{rhel_version} - 1 is available on the given RHEL, 0 - not
        available. it could be the driver is available, but not supported
    comments - the field for comments
    """
    topic = SystemInfoTopic

    pci_id = fields.Nullable(fields.String())
    driver_name = fields.Nullable(fields.String())
    device_name = fields.Nullable(fields.String())
    available_rhel7 = fields.Integer()
    supported_rhel7 = fields.Integer()
    available_rhel8 = fields.Integer()
    supported_rhel8 = fields.Integer()
    available_rhel9 = fields.Integer()
    supported_rhel9 = fields.Integer()
    comment = fields.Nullable(fields.String())


class RestrictedPCIDevices(Model):
    """
    Model to represent all known restricted PCI devices.

    Restricted devices might be restricted based on either
     - pci id (each particular device)
     - driver name (all the family of the devices being served by the
        particular driver).

    However the data type, which represents how the pci id or driver name is
    restricted is identical.

    Thus both attributes has the same data type.

    driver_names - is a container with the devices, which are restricted by
        the driver name identifier
    pci_ids - is a container with the devices, which are restricted by
        the pci_id identifier
    """
    topic = SystemInfoTopic

    driver_names = fields.List(
        fields.Model(RestrictedPCIDevice),
    )
    pci_ids = fields.List(
        fields.Model(RestrictedPCIDevice),
    )



from leapp.actors import Actor
from leapp.libraries.actor.checkmultiplepackageversions import check
from leapp.models import InstalledRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckMultiplePackageVersions(Actor):
    """
    Check for problematic 32bit packages installed together with 64bit ones.

    If a known problematic 32bit package is found, the upgrade will be inhibited with the detailed
    report how to solve the problem if such a remedy exists.
    """

    name = 'multiple_package_versions'
    consumes = (InstalledRPM,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        check()



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import StorageInfo


def inhibit_upgrade_due_var_with_noexec(mountpoint, found_in_fstab=False):
    summary = (
        'Leapp detected that the {0} mountpoint is mounted with the "noexec" option, '
        'which prevents binaries necessary for the upgrade from being executed. '
        'The upgrade process cannot continue with {0} mounted using the "noexec" option.'
    )

    if found_in_fstab:
        hint = (
            'Temporarily remove the "noexec" option from {0} entry in /etc/fstab until the system is upgraded, '
            'and remount the partition without the "noexec" option.'
        )
        related_resource = [reporting.RelatedResource('file', '/etc/fstab')]
    else:
        hint = (
            'Remount {0} without the noexec option and make sure the change is persistent'
            'during the entire in-place upgrade process.'
        )
        related_resource = []

    reporting.create_report([
        reporting.Title(
            'Detected partitions mounted with the "noexec" option, preventing a successful in-place upgrade.'
        ),
        reporting.Summary(summary.format(mountpoint)),
        reporting.Remediation(hint=hint.format(mountpoint)),
        reporting.RelatedResource('file', '/etc/fstab'),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.FILESYSTEM]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
    ] + related_resource)


def find_mount_entry_with_mountpoint(mount_entries, mountpoint):
    for mount_entry in mount_entries:
        if mount_entry.mount == mountpoint:
            return mount_entry
    return None


def find_fstab_entry_with_mountpoint(fstab_entries, mountpoint):
    for fstab_entry in fstab_entries:
        if fstab_entry.fs_file == mountpoint:
            return fstab_entry
    return None


def check_noexec_on_var(storage_info):
    """Check for /var or /var/lib being mounted with noexec mount option."""

    # Order of checking is important as mount options on /var/lib override those on /var
    mountpoints_to_check = ('/var/lib/leapp', '/var/lib', '/var')
    for mountpoint in mountpoints_to_check:
        fstab_entry = find_fstab_entry_with_mountpoint(storage_info.fstab, mountpoint)
        if fstab_entry and 'noexec' in fstab_entry.fs_mntops.split(','):
            inhibit_upgrade_due_var_with_noexec(fstab_entry.fs_file, found_in_fstab=True)
            return  # Do not check further as present mounts would likely reflect fstab

    # Make sure present mountpoints don't contain noexec as well - user might have fixed noexec in fstab
    # but did not remount the partition, or, less likely, mounted the partition without creating a fstab entry
    for mountpoint in mountpoints_to_check:
        mount_entry = find_mount_entry_with_mountpoint(storage_info.mount, mountpoint)
        if mount_entry and 'noexec' in mount_entry.options.split(','):
            inhibit_upgrade_due_var_with_noexec(mount_entry.mount, found_in_fstab=False)
            return


def check_mount_options():
    for storage_info in api.consume(StorageInfo):
        check_noexec_on_var(storage_info)



def global_value(config, default):
    """
    Find the global value for PermitRootLogin option in sshd_config.

    OpenSSH is using the first value found in configuration file, that is not
    in match block other than "all". If there is no such option, the argument
    "default" will be returned.
    """
    for opt in config.permit_root_login:
        if (opt.in_match is None or opt.in_match[0].lower() == 'all'):
            return opt.value
    return default


def semantics_changes(config):
    """
    Check if the current configuration changes semantics if upgraded from RHEL7 to RHEL8

    The case where the configuration does not contain *any* PermitRootLogin option is
    already covered in the actor and does not need to be handled here.

    This tries to capture the case, where the root login is enabled in at least one
    match block. The global default changes so the new configurations will not allow
    all password root logins, but there is at least some chance to access the system as
    root with password.

    Examples:
    * If the root login is globally set (enabled or disabled), the semantics stays the same.
    * If the root login is enabled only in match blocks, the semantics changes, but the
      machine stays accessible at least for clients matching this block.

    """
    config_global_value = global_value(config, None)
    in_match_enabled = False
    if not config.permit_root_login:
        return True

    for opt in config.permit_root_login:
        if opt.value == "yes" and opt.in_match is not None and \
                                  opt.in_match[0].lower() != 'all':
            in_match_enabled = True

    return config_global_value is None and not in_match_enabled



from leapp.models import DistributionSignedRPM, FilteredRpmTransactionTasks, Module, RPM, RpmTransactionTasks
from leapp.snactor.fixture import current_actor_context

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(FilteredRpmTransactionTasks)


def test_actor_execution_with_sample_data(current_actor_context):
    installed_rpm = [
        RPM(name='sample01', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_PGP_SIG'),
        RPM(name='sample02', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='SOME_PGP_SIG')]
    modules_to_enable = [Module(name='enable', stream='1'), Module(name='enable', stream='2')]
    modules_to_reset = [Module(name='reset', stream='1'), Module(name='reset', stream='2')]
    current_actor_context.feed(DistributionSignedRPM(items=installed_rpm))
    current_actor_context.feed(RpmTransactionTasks(
        to_remove=[rpm.name for rpm in installed_rpm],
        to_keep=[installed_rpm[0].name],
        modules_to_enable=modules_to_enable,
        modules_to_reset=modules_to_reset,
    ))
    current_actor_context.feed(RpmTransactionTasks(
        modules_to_enable=modules_to_enable,
        modules_to_reset=modules_to_reset,
    ))
    current_actor_context.run()
    result = current_actor_context.consume(FilteredRpmTransactionTasks)
    assert len(result) == 1
    assert result[0].to_keep == [installed_rpm[0].name]
    assert result[0].to_remove == [installed_rpm[1].name]

    assert len(result[0].modules_to_enable) == 2
    assert all(m.name == 'enable' for m in result[0].modules_to_enable)
    assert '1' in {m.stream for m in result[0].modules_to_enable}
    assert '2' in {m.stream for m in result[0].modules_to_enable}

    assert len(result[0].modules_to_reset) == 2
    assert all(m.name == 'reset' for m in result[0].modules_to_reset)
    assert '1' in {m.stream for m in result[0].modules_to_reset}
    assert '2' in {m.stream for m in result[0].modules_to_reset}



import re

from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import CPUInfo, DetectedDeviceOrDriver, DeviceDriverDeprecationData

LSCPU_NAME_VALUE = re.compile(r'(?P<name>[^:]+):\s+(?P<value>.+)\n?')
PPC64LE_MODEL = re.compile(r'\d+\.\d+ \(pvr (?P<family>[0-9a-fA-F]+) 0*[0-9a-fA-F]+\)')


def _get_lscpu_output():
    try:
        result = run(['lscpu'])
        return result.get('stdout', '')
    except (OSError, CalledProcessError):
        api.current_logger().debug('Executing `lscpu` failed', exc_info=True)
    return ''


def _get_cpu_flags(lscpu):
    flags = lscpu.get('Flags', '')
    return flags.split()


def _get_cpu_entries_for(arch_prefix):
    result = []
    for message in api.consume(DeviceDriverDeprecationData):
        result.extend([
            entry for entry in message.entries
            if entry.device_type == 'cpu' and entry.device_id.startswith(arch_prefix)
        ])
    return result


def _is_detected_aarch64(lscpu, entry):
    # Currently not applicable - We don't have a way to correctly detect this
    # But we should find a way to properly do so in future.

    # Shut up warnings:
    entry = lscpu
    lscpu = entry
    return False


def _is_detected_s390x(lscpu, entry):
    try:
        _, _, machine_type, _ = entry.device_id.split(':', 4)
        return machine_type == lscpu.get('Machine type')
    except ValueError:
        return False


def _is_detected_ppc64le(lscpu, entry):
    try:
        _, _, machine_type, _ = entry.device_id.split(':', 4)
        match = PPC64LE_MODEL.match(lscpu.get('Model'))
        if not match:
            return False
        family = match.group('family')
        return family and machine_type.lstrip('0').lower() == family.lstrip('0').lower()
    except ValueError:
        return False


def _make_set(value):
    """
    Converts range/set specification to a concrete set of numbers

    '[1-3]'         => {1, 2, 3}
    '{1,2,3}'       => {1, 2, 3}
    '{[1-3]}        => {1, 2, 3}
    '{[1-3],[5-7]}  => {1, 2, 3, 5, 6, 7}
    """
    result = set()
    for vrange in value.strip('{} ').split(','):
        if '[' not in vrange:
            try:
                result.add(int(vrange))
            except ValueError:
                pass
        else:
            try:
                start, end = vrange.strip('[] ').split('-')
                result.update(range(int(start.strip()), int(end.strip()) + 1))
            except ValueError:
                pass
    return result


def _match_model(needle, hay):
    try:
        ineedle = int(needle)
    except ValueError:
        return False
    if not hay or hay == '*':
        # Match all
        return True
    if '[' not in hay and '{' not in hay:
        return hay == needle
    return ineedle in _make_set(hay)


def _is_detected_x86_64(lscpu, entry):
    vendors = {'amd': 'AuthenticAMD', 'intel': 'GenuineIntel'}
    try:
        _, vendor, family, model = entry.device_id.split(':', 4)
        if vendor in vendors and vendors[vendor] == lscpu.get('Vendor ID'):
            if family == lscpu.get('CPU family'):
                return _match_model(lscpu.get('Model'), model)
        return False
    except ValueError:
        return False


def _to_detected_device(entry):
    return DetectedDeviceOrDriver(**entry.dump())


def _find_deprecation_data_entries(lscpu):
    arch_prefix, is_detected = None, None
    if architecture.matches_architecture(architecture.ARCH_X86_64):
        arch_prefix, is_detected = architecture.ARCH_X86_64, _is_detected_x86_64
    elif architecture.matches_architecture(architecture.ARCH_PPC64LE):
        arch_prefix, is_detected = architecture.ARCH_PPC64LE, _is_detected_ppc64le
    elif architecture.matches_architecture(architecture.ARCH_S390X):
        arch_prefix, is_detected = architecture.ARCH_S390X, _is_detected_s390x
    elif architecture.matches_architecture(architecture.ARCH_ARM64):
        arch_prefix, is_detected = architecture.ARCH_ARM64, _is_detected_aarch64

    if arch_prefix and is_detected:
        return [
            _to_detected_device(entry) for entry in _get_cpu_entries_for(arch_prefix)
            if is_detected(lscpu, entry)
        ]

    api.current_logger().warning('Unsupported platform could not detect relevant CPU information')
    return []


def process():
    lscpu = dict(LSCPU_NAME_VALUE.findall(_get_lscpu_output()))
    api.produce(*_find_deprecation_data_entries(lscpu))
    # Backwards compatibility
    machine_type = lscpu.get('Machine type')
    flags = _get_cpu_flags(lscpu)
    api.produce(
            CPUInfo(
                machine_type=int(machine_type) if machine_type else None,
                flags=flags
                )
            )



from leapp.actors import Actor
from leapp.libraries.actor import customnetworkscripts
from leapp.models import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class CheckCustomNetworkScripts(Actor):
    """
    Check the existence of custom network-scripts and warn user about possible
    manual intervention requirements.
    """

    name = "check_custom_network_scripts"
    produces = (Report,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        customnetworkscripts.process()



from leapp.actors import Actor
from leapp.libraries.actor import mount_target_iso
from leapp.models import TargetOSInstallationImage, TargetUserSpaceInfo
from leapp.tags import IPUWorkflowTag, PreparationPhaseTag


class MountTargetISO(Actor):
    """Mounts target OS ISO in order to install upgrade packages from it."""

    name = 'mount_target_iso'
    consumes = (TargetUserSpaceInfo, TargetOSInstallationImage,)
    produces = ()
    tags = (PreparationPhaseTag, IPUWorkflowTag)

    def process(self):
        mount_target_iso.mount_target_iso()
